#include <iostream>
#include <fstream>
#include <stdexcept>
#include <vector>
#include <stdint.h>
#include <climits>
#include <cassert>
#include <algorithm>
#include <sys/time.h>

#include "CL/opencl.h"

#include "setupGPU.h"
#include "memCNTRL.h"
#include "utilsPaPaRa.h"

int References =256;
int Queries = 200;


static double gettime(void )
{
    struct timeval ttime;
    gettimeofday(&ttime , 0);
    return ttime.tv_sec + ttime.tv_usec * 0.000001;
}


class timer {
    double m_st;
    
public:
    timer() : m_st(gettime()) {};
    
    double elapsed() const {
        return gettime() - m_st;
    }
    
    
};



typedef int score_t;
typedef unsigned short score_t_exp;
typedef unsigned int u32_de;


struct align_arrays {
    std::vector<score_t> s;
   // std::vector<score_t> si;
};


static score_t align_pvec_score( std::vector<int> &a, std::vector<unsigned int> &a_aux, const std::vector<uint8_t> &b, score_t mismatch_score, score_t match_cgap, score_t gap_open, score_t gap_extend, align_arrays &arr ) {

    if( arr.s.size() < a.size()  ) {
        arr.s.resize( a.size() );
    }
    
    const score_t aux_cgap = 0x1;
    
    const size_t band_width = a.size() - b.size();
    std::fill( arr.s.begin(), arr.s.end(), 0 );
    
    const score_t LARGE = 32000;
//     std::fill( arr.si.begin(), arr.si.end(), LARGE );
    arr.s.at(band_width+1) = LARGE;

    score_t min_score = LARGE;
    
    for( size_t ib = 0; ib < b.size(); ib++ ) {
        int bc = b[ib];
        
        score_t last_sl = LARGE;
        score_t last_sc = LARGE;
        score_t last_sdiag = 0.0;
        
        size_t astart = ib;
                
        score_t * __restrict s_iter = &arr.s[0];

        
        bool lastrow = ib == (b.size() - 1);
        
        last_sdiag = *s_iter;
        
        
        for( size_t ia = astart; ia <= ib + band_width; ++ia, ++s_iter ) {  
            score_t ac = a[ia];
            const bool cgap = a_aux[ia] == aux_cgap;
            
            // determine match or mis-match according to parsimony bits coming from the tree.
            bool is_match = ( ac & bc ) != 0;

            score_t sm = last_sdiag;
            
            if( !is_match ) {
                sm += mismatch_score;
            }
            
            last_sdiag = *(s_iter+1);

            score_t last_sc_OPEN; 
            score_t sl_score_stay;
            
            if( cgap ) {
                last_sc_OPEN = last_sc; 
                sl_score_stay = last_sl;
                sm += match_cgap;
            } else {
                last_sc_OPEN = last_sc + gap_open + gap_extend;  
                sl_score_stay = last_sl + gap_open;
            }
            
            score_t sl = std::min( sl_score_stay, last_sc_OPEN );
                        
            last_sl = sl;
            

            score_t su = last_sdiag + mismatch_score;
            
//             std::cout << "x: " << sm << " " << su << " " << sl << "  ";
            score_t sc = std::min( sm, std::min( su, sl ) );
            
            last_sc = sc;
            *s_iter = sc;
            
        
            if( lastrow ) {
                if( sc < min_score ) {
                    
                    min_score = sc;
                }
            }
           
        }
//         std::cout << "\n";
//         getchar();
    }

    return min_score;
    
}



static score_t align_pvec_scoreBLOCKEDREFERENCE( std::vector<int> &a, std::vector<unsigned int> &a_aux, const std::vector<uint8_t> &b, score_t mismatch_score, score_t match_cgap, score_t gap_open, score_t gap_extend, align_arrays &arr ) {


    int reference_length = a.size();
    int query_length     = b.size();
    int bandwidth = reference_length - query_length;
    //std::cout<<"\n\nReferenceLength:"<<reference_length<<"-Query length:"<<query_length<<"-"<<bandwidth<<"\n";
    int temp_i;  

    int * SC_matrix_line = (int *) malloc (sizeof(int) * reference_length);
    for(temp_i=0;temp_i<reference_length;temp_i++)
    {
	SC_matrix_line[temp_i]=0;
    }
    SC_matrix_line[bandwidth] = 32000;

    int ref_i, que_i;

    int myD, myU, myL, nxtD, nxtU, nxtL, mySL, nxtSL, ref_first, ref_last;
  
    int mySCORE=32000;

    for(que_i=0;que_i<query_length;que_i++)
    {
	
	if(que_i==0) // first row
	{	
		int queVALUE = b[que_i];
				
		ref_first = que_i;
		SC_matrix_line[ref_first+bandwidth ] = 32000;
		ref_last = ref_first + bandwidth + 1;
		
		nxtD = 0;
		nxtL = 32000;
		nxtSL = 32000;
		
		for(ref_i=ref_first;ref_i<ref_last;ref_i++)
		{
			
			myD = nxtD;
			myU = SC_matrix_line[ref_i];
			myL = nxtL;
			mySL = nxtSL;

			int refVALUE = a[ref_i];
			unsigned int refAUX = a_aux[ref_i];
			
			unsigned int CGAP=refAUX;

			unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;

			int sm = myD;
			
			if(match==0)
			{
				sm += mismatch_score;
			}


			int last_sc_open = 0;
			int sl_score_stay = 0;
			
			if(CGAP==1)
			{
				last_sc_open = myL;
				sl_score_stay = mySL;
				sm += match_cgap;
			} 
			else
			{
				last_sc_open = myL + gap_open + gap_extend;
				sl_score_stay = mySL + gap_open;				
			}

			int sl = std::min( sl_score_stay, last_sc_open );

			nxtSL = sl;

			int su = myU + mismatch_score;	

			int sc = std::min( sm, std::min( su, sl ) );

			nxtL = sc;

			SC_matrix_line[ref_i] = sc;

			nxtD = myU;					
			
						
		}		
	}
	else
	{

		int queVALUE = b[que_i];
				
		ref_first = que_i;
		SC_matrix_line[ref_first+bandwidth] = 32000;
		ref_last = ref_first + bandwidth + 1;
		
		nxtD = SC_matrix_line[que_i-1];
		nxtL = 32000;
		nxtSL = 32000;
		
		for(ref_i=ref_first;ref_i<ref_last;ref_i++)
		{
			
			myD = nxtD;
			myU = SC_matrix_line[ref_i];
			myL = nxtL;
			mySL = nxtSL;

			int refVALUE = a[ref_i];
			unsigned int refAUX = a_aux[ref_i];
			
			unsigned int CGAP=refAUX;

			unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;

			int sm = myD;
			
			if(match==0)
			{
				sm += mismatch_score;
			}


			int last_sc_open = 0;
			int sl_score_stay = 0;
			
			if(CGAP==1)
			{
				last_sc_open = myL;
				sl_score_stay = mySL;
				sm += match_cgap;
			} 
			else
			{
				last_sc_open = myL + gap_open + gap_extend;
				sl_score_stay = mySL + gap_open;				
			}

			int sl = std::min( sl_score_stay, last_sc_open );

			nxtSL = sl;

			int su = myU + mismatch_score;	

			int sc = std::min( sm, std::min( su, sl ) );

			nxtL = sc;

			SC_matrix_line[ref_i] = sc;

			nxtD = myU;

			if(que_i==query_length-1)
			{
				mySCORE = std::min(mySCORE,sc);
				
			}					
			
						
		}	
		
	}
	
    }
    

    return mySCORE;//min_score;
    
}


static score_t align_pvec_scoreBLOCKED_OPTIMIZED( std::vector<int> &a, std::vector<unsigned int> &a_aux, const std::vector<uint8_t> &b, score_t mismatch_score, score_t match_cgap, score_t gap_open, score_t gap_extend, align_arrays &arr ) {


    int reference_length = a.size();
    int query_length     = b.size();
    int bandwidth = reference_length - query_length;
  
    int temp_i;  

    int * SC_matrix_line = (int *) malloc (sizeof(int) * reference_length);

    int * SC_matrix_column = (int *) malloc (sizeof(int)*query_length);
    int * SL_matrix_column = (int *) malloc (sizeof(int)*query_length);





    for(temp_i=0;temp_i<reference_length;temp_i++)
    {
	SC_matrix_line[temp_i]=0;
    }
    SC_matrix_line[bandwidth] = 32000;

    int ref_i, que_i;

    int myD, myU, myL, nxtD, nxtU, nxtL, mySL, nxtSL, ref_first, ref_last;
  
    int mySCORE=32000;

	
    int blocksize = 200;
    int blockiterations = 0;

    int lastblock_refstart = reference_length - query_length;

    int total = reference_length;
    int partF = query_length; 
    int partL = query_length;

    int partM = total - partF - partL;

    	blockiterations = partM / blocksize;

    int sitesinblocks = blockiterations * blocksize;
    int partLext = partM - sitesinblocks;
    
	partL = partL + partLext;

	lastblock_refstart = total - partL;















	

	//BLOCK 0

		
	for(que_i=0;que_i<query_length;que_i++)    
    	{	
	
		int queVALUE = b[que_i];
		
		ref_first = que_i;
	
		nxtD = que_i==0?0:SC_matrix_line[que_i-1];

		nxtL = 32000;
		nxtSL = 32000;

		for(ref_i=ref_first;ref_i<query_length;ref_i++)
		{
	
			myD = nxtD;
			myU = SC_matrix_line[ref_i];
			myL = nxtL;
			mySL = nxtSL;

			int refVALUE = a[ref_i];
			unsigned int refAUX = a_aux[ref_i];

			unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;


			int sm = myD ;

			if(match==0)
			{
				sm += mismatch_score;
			}


			int last_sc_open = 0;
			int sl_score_stay = 0;

			if(refAUX==1)
			{
				last_sc_open = myL;
				sl_score_stay = mySL;
				sm += match_cgap;
			} 
			else
			{
				last_sc_open = myL + gap_open + gap_extend;
				sl_score_stay = mySL + gap_open;				
			}

			int sl = std::min( sl_score_stay, last_sc_open );

			nxtSL = sl;

			int su = myU + mismatch_score;	

			int sc = std::min( sm, std::min( su, sl ) );

			nxtL = sc;

			SC_matrix_line[ref_i] = sc;

			nxtD = myU;

			if(que_i==query_length-1)
			{
				mySCORE = std::min(mySCORE,sc);	
			}					
		}
		SL_matrix_column[que_i] = nxtSL;
		SC_matrix_column[que_i] = nxtL;
	}

/*************************************************************************************************/
	
	unsigned int blocki=0;

	for(blocki=0;blocki<blockiterations;blocki++)
	{
		//BLOCK 1
		int nxtD_bu = 0;
		for(que_i=0;que_i<query_length;que_i++)    
	    	{	
	
			int queVALUE = b[que_i];
		
			ref_first = que_i;	
		
			nxtD = nxtD_bu;

			nxtL = SC_matrix_column[que_i];
			nxtSL = SL_matrix_column[que_i];
			nxtD_bu = nxtL;

			for(ref_i=query_length + blocki*blocksize;ref_i<query_length + blocki*blocksize + blocksize;ref_i++)
			{
	
				myD = nxtD;
				myU = SC_matrix_line[ref_i];
				myL = nxtL;
				mySL = nxtSL;

				int refVALUE = a[ref_i];
				unsigned int refAUX = a_aux[ref_i];

				unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;


				int sm = myD ;

				if(match==0)
				{
					sm += mismatch_score;
				}


				int last_sc_open = 0;
				int sl_score_stay = 0;

				if(refAUX==1)
				{
					last_sc_open = myL;
					sl_score_stay = mySL;
					sm += match_cgap;
				} 
				else
				{
					last_sc_open = myL + gap_open + gap_extend;
					sl_score_stay = mySL + gap_open;				
				}

				int sl = std::min( sl_score_stay, last_sc_open );

				nxtSL = sl;

				//SL_matrix_column[que_i] = sl;

				int su = myU + mismatch_score;	

				int sc = std::min( sm, std::min( su, sl ) );

				nxtL = sc;

				//SC_matrix_column[que_i] = sc;
		
				SC_matrix_line[ref_i] = sc;

				nxtD = myU;

				if(que_i==query_length-1)
				{
					mySCORE = std::min(mySCORE,sc);	
				}					
			}
			SL_matrix_column[que_i] = nxtSL;
			SC_matrix_column[que_i] = nxtL;
		}
	}
/*************************************************************************************************/

	//LAST BLOCK 
	int nxtD_bu = 0;
	for(que_i=0;que_i<query_length;que_i++)    
    	{	
	
		int queVALUE = b[que_i];
		
		SC_matrix_line[que_i+bandwidth] = 32000;
		ref_last = que_i + bandwidth + 1;
		
		nxtD = nxtD_bu;

		nxtL = SC_matrix_column[que_i];
		nxtSL = SL_matrix_column[que_i];
		nxtD_bu = nxtL;

		for(ref_i=lastblock_refstart;ref_i<ref_last;ref_i++)
		{
	
			myD = nxtD;
			myU = SC_matrix_line[ref_i];
			myL = nxtL;
			mySL = nxtSL;

			int refVALUE = a[ref_i];
			unsigned int refAUX = a_aux[ref_i];

			unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;

			int sm = myD ;

			if(match==0)
			{
				sm += mismatch_score;
			}


			int last_sc_open = 0;
			int sl_score_stay = 0;

			if(refAUX==1)
			{
				last_sc_open = myL;
				sl_score_stay = mySL;
				sm += match_cgap;
			} 
			else
			{
				last_sc_open = myL + gap_open + gap_extend;
				sl_score_stay = mySL + gap_open;				
			}

			int sl = std::min( sl_score_stay, last_sc_open );

			nxtSL = sl;

			int su = myU + mismatch_score;	

			int sc = std::min( sm, std::min( su, sl ) );

			nxtL = sc;
		
			SC_matrix_line[ref_i] = sc;

			nxtD = myU;

			if(que_i==query_length-1)
			{
				mySCORE = std::min(mySCORE,sc);	
			}					
		}
	}

















    return mySCORE;    
}




static score_t align_pvec_scoreBLOCKED_OPTIMIZED_2( std::vector<int> &a, std::vector<unsigned int> &a_aux, const std::vector<uint8_t> &b, score_t mismatch_score, score_t match_cgap, score_t gap_open, score_t gap_extend, align_arrays &arr ) {


    int reference_length = a.size();
    int query_length     = b.size();
    int bandwidth = reference_length - query_length;
  
    int temp_i;  

    int * SC_matrix_line = (int *) malloc (sizeof(int) * reference_length);

    int * SC_matrix_column = (int *) malloc (sizeof(int)*query_length);
    int * SL_matrix_column = (int *) malloc (sizeof(int)*query_length);





    for(temp_i=0;temp_i<reference_length;temp_i++)
    {
	SC_matrix_line[temp_i]=0;
    }
    SC_matrix_line[bandwidth] = 32000;

    int ref_i, que_i;

    int myD, myU, myL, nxtD, nxtU, nxtL, mySL, nxtSL, ref_first, ref_last;
  
    int mySCORE=32000;

	
    int blocksize = 80;
    int blockiterations = 0;

    int lastblock_refstart = reference_length - query_length;

    int total = reference_length;
    int partF = query_length; 
    int partL = query_length;

    int partM = total - partF - partL;

    	blockiterations = partM / blocksize;

    int sitesinblocks = blockiterations * blocksize;
    int partLext = partM - sitesinblocks;
    
	partL = partL + partLext;

	lastblock_refstart = total - partL;












	int * SC_matrix_line_first = (int *) malloc (sizeof(int) * query_length);
	

	//BLOCK 0

		
	for(que_i=0;que_i<query_length;que_i++)    
    	{	
	
		int queVALUE = b[que_i];
		
		ref_first = que_i;
	
		nxtD = que_i==0?0:SC_matrix_line_first[que_i-1];

		nxtL = 32000;
		nxtSL = 32000;

		for(ref_i=ref_first;ref_i<query_length;ref_i++)
		{
	
			myD = nxtD;
			myU = que_i==0?0:SC_matrix_line_first[ref_i];
			myL = nxtL;
			mySL = nxtSL;

			int refVALUE = a[ref_i];
			unsigned int refAUX = a_aux[ref_i];

			unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;


			int sm = myD ;

			if(match==0)
			{
				sm += mismatch_score;
			}


			int last_sc_open = 0;
			int sl_score_stay = 0;

			if(refAUX==1)
			{
				last_sc_open = myL;
				sl_score_stay = mySL;
				sm += match_cgap;
			} 
			else
			{
				last_sc_open = myL + gap_open + gap_extend;
				sl_score_stay = mySL + gap_open;				
			}

			int sl = std::min( sl_score_stay, last_sc_open );

			nxtSL = sl;

			int su = myU + mismatch_score;	

			int sc = std::min( sm, std::min( su, sl ) );

			nxtL = sc;

			SC_matrix_line_first[ref_i] = sc;

			nxtD = myU;

			if(que_i==query_length-1)
			{
				mySCORE = std::min(mySCORE,sc);	
			}					
		}
		SL_matrix_column[que_i] = nxtSL;
		SC_matrix_column[que_i] = nxtL;
	}

/*************************************************************************************************/
	
	unsigned int blocki=0;

	int * SC_matrix_line_block = (int *) malloc (sizeof(int) * blocksize);

	for(blocki=0;blocki<blockiterations;blocki++)
	{
		//BLOCK 1
		int nxtD_bu = 0;
		for(que_i=0;que_i<query_length;que_i++)    
	    	{	
	
			int queVALUE = b[que_i];
		
			ref_first = que_i;	
		
			nxtD = nxtD_bu;

			nxtL = SC_matrix_column[que_i];
			nxtSL = SL_matrix_column[que_i];
			nxtD_bu = nxtL;
			unsigned int internalindex=0;

			for(ref_i=query_length + blocki*blocksize;ref_i<query_length + blocki*blocksize + blocksize;ref_i++)
			{
	
				myD = nxtD;
				myU = que_i==0?0:SC_matrix_line_block[internalindex];
				myL = nxtL;
				mySL = nxtSL;

				int refVALUE = a[ref_i];
				unsigned int refAUX = a_aux[ref_i];

				unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;


				int sm = myD ;

				if(match==0)
				{
					sm += mismatch_score;
				}


				int last_sc_open = 0;
				int sl_score_stay = 0;

				if(refAUX==1)
				{
					last_sc_open = myL;
					sl_score_stay = mySL;
					sm += match_cgap;
				} 
				else
				{
					last_sc_open = myL + gap_open + gap_extend;
					sl_score_stay = mySL + gap_open;				
				}

				int sl = std::min( sl_score_stay, last_sc_open );

				nxtSL = sl;

				//SL_matrix_column[que_i] = sl;

				int su = myU + mismatch_score;	

				int sc = std::min( sm, std::min( su, sl ) );

				nxtL = sc;

				//SC_matrix_column[que_i] = sc;
		
				SC_matrix_line_block[internalindex] = sc;

				nxtD = myU;

				if(que_i==query_length-1)
				{
					mySCORE = std::min(mySCORE,sc);	
				}				
				internalindex++;	
			}
			SL_matrix_column[que_i] = nxtSL;
			SC_matrix_column[que_i] = nxtL;
		}				
			
	}
/*************************************************************************************************/

	//LAST BLOCK 
	int nxtD_bu = 0;

	int * SC_matrix_line_last = (int *) malloc (sizeof(int) * partL);


	for(que_i=0;que_i<query_length;que_i++)    
    	{	
	
		int queVALUE = b[que_i];
		
		//SC_matrix_line[que_i+bandwidth] = 32000;
		//SC_matrix_line_last[partL]=32000;
		ref_last = que_i + bandwidth + 1;
		
		nxtD = nxtD_bu;

		nxtL = SC_matrix_column[que_i];
		nxtSL = SL_matrix_column[que_i];
		nxtD_bu = nxtL;

		int internalindex=0;
		for(ref_i=lastblock_refstart;ref_i<ref_last;ref_i++)
		{
	
			myD = nxtD;
			myU = que_i==0?0:SC_matrix_line_last[internalindex];
			myU = ref_i==ref_last-1?32000:myU;
			myL = nxtL;
			mySL = nxtSL;

			int refVALUE = a[ref_i];
			unsigned int refAUX = a_aux[ref_i];

			unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;

			int sm = myD ;

			if(match==0)
			{
				sm += mismatch_score;
			}


			int last_sc_open = 0;
			int sl_score_stay = 0;

			if(refAUX==1)
			{
				last_sc_open = myL;
				sl_score_stay = mySL;
				sm += match_cgap;
			} 
			else
			{
				last_sc_open = myL + gap_open + gap_extend;
				sl_score_stay = mySL + gap_open;				
			}

			int sl = std::min( sl_score_stay, last_sc_open );

			nxtSL = sl;

			int su = myU + mismatch_score;	

			int sc = std::min( sm, std::min( su, sl ) );

			nxtL = sc;
		
			SC_matrix_line_last[internalindex] = sc;

			nxtD = myU;

			if(que_i==query_length-1)
			{
				mySCORE = std::min(mySCORE,sc);	
			}
			internalindex++;					
		}
	}

















    return mySCORE;    
}

static score_t align_pvec_scoreBLOCKED_OPTIMIZED_3( std::vector<int> &a, std::vector<unsigned int> &a_aux, const std::vector<uint8_t> &b, score_t mismatch_score, score_t match_cgap, score_t gap_open, score_t gap_extend, align_arrays &arr ) {


    int reference_length = a.size();
    int query_length     = b.size();
    int bandwidth = reference_length - query_length;
  
    int temp_i;  

   
    int * SC_matrix_column = (int *) malloc (sizeof(int)*query_length);
    int * SL_matrix_column = (int *) malloc (sizeof(int)*query_length);





   
    int ref_i, que_i;

    int myD, myU, myL, nxtD, nxtU, nxtL, mySL, nxtSL, ref_first, ref_last;
  
    int mySCORE=32000;

	
    int blocksize = 20;
    int blockiterations = 0;

    int lastblock_refstart = reference_length - query_length;

    int total = reference_length;
    int partF = query_length; 
    int partL = query_length;

    int partM = total - partF - partL;

    	blockiterations = partM / blocksize;

    int sitesinblocks = blockiterations * blocksize;
    int partLext = partM - sitesinblocks;
    
	partL = partL + partLext;

	lastblock_refstart = total - partL;


	//std::cout<<"\n"<<blockiterations<<"-"<<lastblock_refstart;









	int * SC_matrix_line_first = (int *) malloc (sizeof(int) * (125 + blocksize)); // 1. Query length
	

	//BLOCK 0

		
	for(que_i=0;que_i<query_length;que_i++)    
    	{	
	
		int queVALUE = b[que_i];
		
		ref_first = que_i;
	
		nxtD = que_i==0?0:SC_matrix_line_first[que_i-1];

		nxtL = 32000;
		nxtSL = 32000;

		for(ref_i=ref_first;ref_i<query_length;ref_i++)
		{
	
			myD = nxtD;
			myU = que_i==0?0:SC_matrix_line_first[ref_i];
			myL = nxtL;
			mySL = nxtSL;

			int refVALUE = a[ref_i];
			unsigned int refAUX = a_aux[ref_i];

			unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;


			int sm = myD ;

			if(match==0)
			{
				sm += mismatch_score;
			}


			int last_sc_open = 0;
			int sl_score_stay = 0;

			if(refAUX==1)
			{
				last_sc_open = myL;
				sl_score_stay = mySL;
				sm += match_cgap;
			} 
			else
			{
				last_sc_open = myL + gap_open + gap_extend;
				sl_score_stay = mySL + gap_open;				
			}

			int sl = std::min( sl_score_stay, last_sc_open );

			nxtSL = sl;

			int su = myU + mismatch_score;	

			int sc = std::min( sm, std::min( su, sl ) );

			nxtL = sc;

			SC_matrix_line_first[ref_i] = sc;

			nxtD = myU;

			if(que_i==query_length-1)
			{
				mySCORE = std::min(mySCORE,sc);	
			}					
		}
		SL_matrix_column[que_i] = nxtSL;
		SC_matrix_column[que_i] = nxtL;
	}

/*************************************************************************************************/
	
	unsigned int blocki=0;

	int * SC_matrix_line_block = (int *) malloc (sizeof(int) * blocksize); // 2. blocksize

	for(blocki=0;blocki<blockiterations;blocki++) // 3. blockiterations
	{
		//BLOCK 1
		int nxtD_bu = 0;
		for(que_i=0;que_i<query_length;que_i++)    
	    	{	
	
			int queVALUE = b[que_i];
		
			ref_first = que_i;	
		
			nxtD = nxtD_bu;

			nxtL = SC_matrix_column[que_i];
			nxtSL = SL_matrix_column[que_i];
			nxtD_bu = nxtL;
			unsigned int internalindex=0;

			for(ref_i=query_length + blocki*blocksize;ref_i<query_length + blocki*blocksize + blocksize;ref_i++)
			{
	
				myD = nxtD;
				myU = que_i==0?0:SC_matrix_line_block[internalindex];
				myL = nxtL;
				mySL = nxtSL;

				int refVALUE = a[ref_i];
				unsigned int refAUX = a_aux[ref_i];

				unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;


				int sm = myD ;

				if(match==0)
				{
					sm += mismatch_score;
				}


				int last_sc_open = 0;
				int sl_score_stay = 0;

				if(refAUX==1)
				{
					last_sc_open = myL;
					sl_score_stay = mySL;
					sm += match_cgap;
				} 
				else
				{
					last_sc_open = myL + gap_open + gap_extend;
					sl_score_stay = mySL + gap_open;				
				}

				int sl = std::min( sl_score_stay, last_sc_open );

				nxtSL = sl;

				//SL_matrix_column[que_i] = sl;

				int su = myU + mismatch_score;	

				int sc = std::min( sm, std::min( su, sl ) );

				nxtL = sc;

				//SC_matrix_column[que_i] = sc;
		
				SC_matrix_line_block[internalindex] = sc;

				nxtD = myU;

				if(que_i==query_length-1)
				{
					mySCORE = std::min(mySCORE,sc);	
				}				
				internalindex++;	
			}
			SL_matrix_column[que_i] = nxtSL;
			SC_matrix_column[que_i] = nxtL;
		}				
			
	}
/*************************************************************************************************/

	//LAST BLOCK 
	int nxtD_bu = 0;

//	int * SC_matrix_line_last = (int *) malloc (sizeof(int) * partL); // 4. partL


	for(que_i=0;que_i<query_length;que_i++)    
    	{	
	
		int queVALUE = b[que_i];
		
		//SC_matrix_line[que_i+bandwidth] = 32000;
		//SC_matrix_line_last[partL]=32000;
		ref_last = que_i + bandwidth + 1;
		
		nxtD = nxtD_bu;

		nxtL = SC_matrix_column[que_i];
		nxtSL = SL_matrix_column[que_i];
		nxtD_bu = nxtL;

		int internalindex=0;
		for(ref_i=lastblock_refstart;ref_i<ref_last;ref_i++)
		{
	
			myD = nxtD;
			myU = que_i==0?0:SC_matrix_line_first[internalindex];
			myU = ref_i==ref_last-1?32000:myU;
			myL = nxtL;
			mySL = nxtSL;

			int refVALUE = a[ref_i];
			unsigned int refAUX = a_aux[ref_i];

			unsigned int match = ( queVALUE & refVALUE ) != 0 ? 1:0;

			int sm = myD ;

			if(match==0)
			{
				sm += mismatch_score;
			}


			int last_sc_open = 0;
			int sl_score_stay = 0;

			if(refAUX==1)
			{
				last_sc_open = myL;
				sl_score_stay = mySL;
				sm += match_cgap;
			} 
			else
			{
				last_sc_open = myL + gap_open + gap_extend;
				sl_score_stay = mySL + gap_open;				
			}

			int sl = std::min( sl_score_stay, last_sc_open );

			nxtSL = sl;

			int su = myU + mismatch_score;	

			int sc = std::min( sm, std::min( su, sl ) );

			nxtL = sc;
		
			SC_matrix_line_first[internalindex] = sc;

			nxtD = myU;

			if(que_i==query_length-1)
			{
				mySCORE = std::min(mySCORE,sc);	
			}
			internalindex++;					
		}
	}

















    return mySCORE;    
}


class testbench {
    std::vector<std::vector <int> > m_ref_pvecs;
    std::vector<std::vector <unsigned int> > m_ref_aux;
    
    std::vector<std::vector <uint8_t> > m_qs_pvecs;

    
    void load_qs( std::istream &is ) {
        size_t num_qs;
        is >> num_qs;
        
        m_qs_pvecs.resize( num_qs );
        
        for( int i = 0; i < m_qs_pvecs.size(); ++i ) {
            size_t len;
            is >> len;
            while( isspace( is.get() ) && !is.eof() ) {}
            is.unget();
            
            if( is.eof() ) {
                throw std::runtime_error( "unexpected end of file" );
            }
            
            m_qs_pvecs[i].resize(len);
            is.read( (char*)&m_qs_pvecs[i][0], len );
//             std::cout << "qs: " << i << " " << m_qs_pvecs[i].size() << "\n";
        }
    }
    
    void load_ref( std::ifstream &is ) {
        size_t num_ref;
        is >> num_ref;
        m_ref_pvecs.resize( num_ref );
        m_ref_aux.resize( num_ref );
        
        for( int i = 0; i < m_ref_pvecs.size(); ++i ) {
            size_t len;
            is >> len;
            while( isspace( is.get() ) && !is.eof() ) {}
            is.unget();
            
            if( is.eof() ) {
                throw std::runtime_error( "unexpected end of file" );
            }
            
            m_ref_pvecs[i].resize(len);
            is.read( (char*)&m_ref_pvecs[i][0], len * sizeof(int) );
            
            m_ref_aux[i].resize(len);
            is.read( (char*)&m_ref_aux[i][0], len * sizeof(unsigned int) );
            
//             std::cout << "ref: " << i << " " << m_ref_pvecs[i].size() << "\n";
        }
        
    }
    template<typename T, size_t min_len>
    static void trim_ref( std::vector<T> &v ) {
        
    	const size_t max_len = min_len;
        while( v.size() < min_len ) {
            std::vector<T> v_new = v;
            
            
            std::copy( v.begin(), v.end(), std::back_inserter(v_new) );
            
            v.swap(v_new);
            
        }
        

        if( v.size() > max_len ) {
            v.resize( max_len );
	    //v.erase( v.begin(), v.begin() + v.size() - max_len );
        }
        
        
    }   
 
    template<typename cont>
    void duplicate( cont &c ) {
	assert( !c.empty() );
    	cont tmp = c;
	c.insert( c.end(), tmp.begin(), tmp.end() );
    }

public:  
    testbench( const char *qs_name, const char *ref_name ) {
        {
            std::ifstream is( qs_name );
            load_qs( is );
        }
        {
            std::ifstream is( ref_name );
            load_ref( is );
        }

	while( m_ref_pvecs.size() < References ) {
		duplicate( m_ref_pvecs );
		duplicate( m_ref_aux );

	}

	std::srand(1234);
	std::random_shuffle( m_ref_pvecs.begin(), m_ref_pvecs.end() );
	std::srand(1234);
	std::random_shuffle( m_ref_aux.begin(), m_ref_aux.end() );
        
        
        // reduce size of testset
        size_t num_ref = References; // JUMP
        size_t num_qs = Queries;
        
        num_ref = std::min( num_ref, m_ref_pvecs.size() );
        num_qs = std::min( num_qs, m_qs_pvecs.size() );
        
        m_ref_pvecs.resize(num_ref);
        m_ref_aux.resize(num_ref);
        m_qs_pvecs.resize(num_qs);
        
	//const size_t len = 1200;
	//std::for_each( m_ref_pvecs.begin(), m_ref_pvecs.end(), &trim_ref<int,len> );
        //std::for_each( m_ref_aux.begin(), m_ref_aux.end(), &trim_ref<unsigned int,len> );
	
        int ct = 0;
	for( size_t i = 0; i < m_ref_aux.size(); ++i ) {
 	if( m_ref_aux[i].back() != 0 ) { ct++; }
	}
	std::cout << "\n\nCGAP REFS:\t\t"<<ct << "\n";


    }
    
    
    void run_seq() {
        const int score_gap_open = 1;
        const int score_gap_extend = 1;
        const int score_mismatch = 1;
        const int score_match_cgap = 4;
        
        
        
        uint64_t ncup = 0;
        
        
        /******************************************/
	/* Acceleration attempt implemented by me */
        /******************************************/

        /* Setting up GPU.. */

	int i;
	
	cl_int ciErrNum;

   	cl_platform_id* myPlatform = select_NVIDIA_platform();

    	cl_device_id * myDevice = select_GPU_device(myPlatform);

    	cl_context myContext= create_context(myDevice);

    	cl_command_queue myCommandQueue = create_command_queue (myDevice, myContext);

    	int kernel_size;

	
	/* Currently working on *
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Reference.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLReference");
	/* */

 	/* Optimized Kernel 0*
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Optimized0.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLOptimized0");
	/* */

	/* Optimized Kernel 1*
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Optimized1.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLOptimized1");
	/* */

	/* Optimized Kernel 2* NOT WORKING				
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Optimized2.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLOptimized2");
	/* */

	/* Optimized Kernel 3* NOT WORKING
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Optimized3.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLOptimized3");
	/* */

	/* Optimized Kernel 4* NOT WORKING
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Optimized4.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLOptimized4");
	/* */

	/* Optimized Kernel 5*
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Optimized5.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLOptimized5");
	/* */

	/* Optimized Kernel 6*
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Optimized6.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLOptimized6");
	/* */

	/* Optimized Kernel 7*
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Optimized7.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLOptimized7");
	/* */

	/* Optimized Kernel 8* // Working blocked version.. 15% faster than optimized 1
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Optimized8.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLOptimized8");
	/* */

	/* Generic Kernel *
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Generic-Kernel.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLGenericKernel");
	/* */

	/* Compressed Kernel 6*
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Unfold6-Kernel-Compressed.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLUnfold6KernelCompressed");
	/* */

	/* Compressed Kernel 8 *
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Unfold8-Kernel-Compressed.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLUnfold8KernelCompressed");
	/* */

	/* Compressed Kernel 10*
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Unfold10-Kernel-Compressed.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLUnfold10KernelCompressed");
	/* */

	/* Compressed Kernel 12*/
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Unfold12-Kernel-Compressed.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLUnfold12KernelCompressed");
	/* */

	/* Compressed Kernel 12 REV*
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Unfold12-Kernel-Compressed-Reversed.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLUnfold12KernelCompressedReversed");
	/* */


    	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_GLOBAL_MEM.cl", &kernel_size);
        //char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF2.cl", &kernel_size); // FASTEST
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_SHARED_MEM.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_SHARED_MEM2.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_BLOCKED.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_BLOCKED3.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_BLOCKED_TEST.cl", &kernel_size);

    	

    	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_GLOBAL_MEM");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF2");  // FASTEST
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_SHARED_MEM");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_SHARED_MEM2");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_BLOCKED");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_BLOCKED3");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_BLOCKED_TEST");

        /* Memory reorganization 

	const size_t reflen = m_ref_pvecs[m_ref_pvecs.size()-1].size();

	//printf("\n\nReference length (characters): %d\n",reflen);

	const size_t compactRefMode = 1; // # of Character+cgap sets in u32
        const size_t alignBlockSize = 32;
	const size_t compactRefSize = getCompactRefSize (compactRefMode, reflen, alignBlockSize);
	
	//printf("\n\nCompact Size per Ref (u32): %d\n",compactRefSize);

	unsigned int * compactRef = getCompactReference (compactRefMode, reflen, m_ref_pvecs[0], m_ref_aux[0], compactRefSize);
	

	// Simon reference test

	timer SimonTimer;	
	int score_s;
	ncup=0; 
	for(size_t i=0;i<1;i++)
	{
	score_s = align_pvec_score( m_ref_pvecs[0], m_ref_aux[0], m_qs_pvecs[0], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
	 ncup += reflen * m_qs_pvecs[1].size();
	}
	//std::cout << "\n\nSimon score for 0 0 0     " << score_s << " in "<<SimonTimer.elapsed()<<"\n";
	//std::cerr << "GNcups: " << ncup / (SimonTimer.elapsed() * 1e9) << "\n";

 	
	// My experiment




	// Find Maximum Query Length
	//unsigned int maxQuerySize = 0; 	for(size_t i=0;i<m_qs_pvecs.size();i++) {if(maxQuerySize<m_qs_pvecs[i].size()){maxQuerySize=m_qs_pvecs[i].size();}}



/*******************************************
// Load Check
timer GPUTimer;
const size_t num_of_queries = m_qs_pvecs.size();
const size_t num_of_references = m_ref_pvecs.size();

const size_t reference_size = m_ref_pvecs[0].size();
unsigned int maxQuerySize = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(maxQuerySize<m_qs_pvecs[i].size()){maxQuerySize=m_qs_pvecs[i].size();}}
const size_t query_size_max = maxQuerySize;

//std::cout<< num_of_queries << " queries to be aligned to " << num_of_references << " references\n";
//std::cout<< query_size_max << " query length to be aligned to " << reference_size << " reference length\n";


const size_t memBlockSize = 32;
const unsigned int constQueryLength = get_constSequenceLength (query_size_max,memBlockSize,1);
const unsigned int constReferenceLength = get_constSequenceLength (reference_size,memBlockSize,1);



unsigned int referencesAllSize = constReferenceLength * num_of_references;
unsigned int * referencesAll = (unsigned int *) malloc(sizeof(unsigned int)*referencesAllSize);
for(size_t i=0;i<num_of_references;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],constReferenceLength,referencesAll);}

unsigned int queriesAllSize = constQueryLength * num_of_queries;
unsigned int * queriesAll = (unsigned int *) malloc(sizeof(unsigned int)*queriesAllSize);
for(size_t i=0;i<num_of_queries;i++){putSequenceQue(i,m_qs_pvecs[i],constQueryLength,queriesAll);}


cl_mem referencesAllPinned = allocatePinnedHostMemory_u32 (referencesAllSize, myContext, myCommandQueue);
referencesAllPinned = initializePinnedHostMemory_u32_Sequence (referencesAllSize, referencesAllPinned, myCommandQueue, referencesAll);
cl_mem referencesAllDevice = allocateDeviceMemory_u32 (referencesAllSize, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (referencesAllSize, referencesAllPinned, referencesAllDevice, myCommandQueue);
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &referencesAllDevice);

cl_mem queriesAllPinned = allocatePinnedHostMemory_u32 (queriesAllSize, myContext, myCommandQueue);
queriesAllPinned = initializePinnedHostMemory_u32_Sequence (queriesAllSize, queriesAllPinned, myCommandQueue, queriesAll);
cl_mem queriesAllDevice = allocateDeviceMemory_u32 (queriesAllSize, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (queriesAllSize, queriesAllPinned, queriesAllDevice, myCommandQueue);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &queriesAllDevice);




//std::cout<< constReferenceLength << " FIXED query length to be aligned to " << referencesAllSize << " reference length ALL\n";



//size_t testvalue = shrRoundUp((int)memBlockSize, 45);
//std::cout<<"the test value is:"<<testvalue<<"\n";

//FILE * fp;
//fp = fopen("test.txt","w");

//for(size_t i=0;i<queriesAllSize;i++)
//{
//	fprintf(fp,"%u-%u:%u %u\n",i,queriesAll[i], m_qs_pvecs[1][i],m_qs_pvecs[1].size());
//}

//fclose(fp);



//float sharedMemLineSize = get_shareMemLineSize (constReferenceLength);
//std::cout<< sharedMemLineSize << " Shared Memory Required \n";

/*******************************************/


/*** DUMB IMPLEMENTATION FOR GPU TESTING***

//for(size_t i=0;i<20;i++)
//{

const size_t argumentsSize = 32;
const size_t argmult = 1;
const size_t compReferenceMode = 1;
const size_t compQueryMode = 1;
const size_t alignMemBlockSize = 32;

const size_t test1 = 2;
const size_t test2 = 24;


const size_t referenceLength = m_ref_pvecs[m_ref_pvecs.size()-1].size();
const size_t compReferenceLength = getCompactSize (compReferenceMode, referenceLength, alignMemBlockSize);
const size_t queryLength = m_qs_pvecs[0].size();
//std::cout<<"Testing Reference Length:"<<compReferenceLength<<" and Query Length: "<< queryLength<<"\n\n";
	
unsigned int * compactReference = getCompactReference (compReferenceMode, referenceLength, m_ref_pvecs[0], m_ref_aux[0], compReferenceLength);
unsigned int * Query = getCompactQuery(compQueryMode, queryLength, m_qs_pvecs[0], queryLength);


cl_mem referencePinned = allocatePinnedHostMemory_u32 (compReferenceLength, myContext, myCommandQueue);
cl_mem queryPinned = allocatePinnedHostMemory_u32 (queryLength, myContext, myCommandQueue);
cl_mem argumentsPinned = allocatePinnedHostMemory_u32 (test1*test2, myContext, myCommandQueue);


referencePinned = initializePinnedHostMemory_u32_Sequence (compReferenceLength, referencePinned, myCommandQueue, compactReference);						
queryPinned = initializePinnedHostMemory_u32_Sequence (queryLength, queryPinned, myCommandQueue,Query);
argumentsPinned = initializePinnedHostMemory_u32_specialArguments (test1*test2, argumentsPinned, myCommandQueue,argumentsSize*argmult);

cl_mem referenceDevice = allocateDeviceMemory_u32 (compReferenceLength, myContext);
cl_mem queryDevice = allocateDeviceMemory_u32 (queryLength, myContext);
cl_mem argumentsDevice = allocateDeviceMemory_u32 (test1*test2, myContext);
cl_mem tempLineDevice = allocateDeviceMemory_u32 (referencesAllSize, myContext);


copyPinnedHostMemoryToDeviceMemory_u32 (compReferenceLength, referencePinned, referenceDevice, myCommandQueue);
copyPinnedHostMemoryToDeviceMemory_u32 (queryLength, queryPinned, queryDevice, myCommandQueue);
copyPinnedHostMemoryToDeviceMemory_u32 (test1*test2, argumentsPinned, argumentsDevice, myCommandQueue);

//clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &referenceDevice);
//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &queryDevice);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &argumentsDevice);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &tempLineDevice);

//clSetKernelArg(myKernel, 3, sizeof(unsigned int) * constReferenceLength, 0);

clFinish(myCommandQueue);

//size_t workGroupSize[2]; 
 
size_t workGroupSize[] = {test1*test2};
size_t workGroupSize2[] = {test2};//argumentsSize};
//cl_event event;
//clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, workGroupSize, workGroupSize2, 0, NULL, &event);

//clFinish(myCommandQueue);

unsigned int * sequenceReturned = copyDeviceMemoryToPinnedHostMemory_u32 (test1*test2, argumentsDevice, argumentsPinned, myCommandQueue);

clFinish(myCommandQueue);

//for(size_t yy=0; yy < test1*test2; yy++)
//{
//	printf("%d:%u\n",yy,sequenceReturned[yy]);	

//}
//std::cout << "\n\nGPU score for 0 0 0    in "<<GPUTimer.elapsed()<<"\n";


size_t workgroupsize;
clGetKernelWorkGroupInfo (myKernel, *myDevice, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &workgroupsize,NULL);
//std::cout<<"Work Group Size:"<<workgroupsize<<"\n";

cl_ulong localmemsize;
clGetKernelWorkGroupInfo (myKernel, *myDevice, CL_KERNEL_LOCAL_MEM_SIZE, sizeof(cl_ulong), &localmemsize,NULL);
//std::cout<<"Local Mem Size:"<<localmemsize<<"\n";

/***************/

/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION: NO OPTIMIZATIONS 

timer NikosTimer;
const size_t RefNum = 2;
const size_t QueNum = 200;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 2; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing
ArgVec[0] = RefLen;
ArgVec[1] = QueNum;
for(size_t i = 2; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-2].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t ResVecLen = QueNum * RefNum;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &RefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &IntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeLocal[] = {RefNum};//argumentsSize};

cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

//for(size_t i=0;i<ResVecLen;i++)
//{
//	std::cout<<i<<":"<<ResVec[i]<<"\n";
//}
std::cout << "GPU Time: " << NikosTimer.elapsed();

/**/

/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-REFERENCE GLOBAL 

timer NikosTimer;
const size_t RefNum = 192; 
const size_t QueNum = 200;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = RefNum;//get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen;
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}

// SIZE CHECK AND FIT TO SHARED AND GLOBAL MEMORY // TODO IN NEXT STEP
//const size_t SharedMemSize = 48; // KBytes
//const size_t gMemSAddr = get_sGloIndex (SharedMemSize, RefPadBloLen*3);
//const size_t SharedMemSizePerMP = gMemSAddr;
//const size_t GlobalMemSizePerMP = InterRefVecLen - SharedMemSizePerMP;
//const size_t TotalGlobalMemSize = QueNum * GlobalMemSizePerMP;

//std::cout << "Total/ Global/ Shared"<< InterRefVecLen <<"/"<< GlobalMemSizePerMP <<"/" << SharedMemSizePerMP <<"\n";


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
//ArgVec[0] = RefLen;
//ArgVec[1] = QueNum;
//ArgVec[2] = RefPadBloLen;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;

//std::cout<<"Before/After:"<<IntVecLen<<"/"<<InterIntVecLen*4/1024/1024<<"\n";

const size_t ResVecLen = QueNum * RefPadBloLen;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &RefVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
//clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &IntVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);
double NikosTime= NikosTimer.elapsed();
std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/


/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-REFERENCE GLOBAL  2 -- WORKING DONE KERNEL NAME: INTER REF2

timer NikosTimer;
const size_t RefNum = 102;  
const size_t QueNum = 500;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = RefNum;//get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen;
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}

// SIZE CHECK AND FIT TO SHARED AND GLOBAL MEMORY // TODO IN NEXT STEP
const size_t SharedMemSize32 = 48; // KBytes
//const size_t gMemSAddr = get_sGloIndex (SharedMemSize, RefPadBloLen*3);
//const size_t SharedMemSizePerMP = gMemSAddr;
//const size_t GlobalMemSizePerMP = InterRefVecLen - SharedMemSizePerMP;
//const size_t TotalGlobalMemSize = QueNum * GlobalMemSizePerMP;

//std::cout << "Total/ Global/ Shared"<< InterRefVecLen <<"/"<< GlobalMemSizePerMP <<"/" << SharedMemSizePerMP <<"\n";


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
//ArgVec[0] = RefLen;
//ArgVec[1] = QueNum;
//ArgVec[2] = RefPadBloLen;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;

//std::cout<<"Before/After:"<<IntVecLen<<"/"<<InterIntVecLen*4/1024/1024<<"\n";

const size_t ResVecLen = QueNum * RefPadBloLen;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &RefVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
//clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &IntVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);
double NikosTime= NikosTimer.elapsed();
std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/


/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-QUERY GLOBAL -- KERNEL NAME: INTER QUE 

timer NikosTimer;
const size_t RefNum = 32; // JUMP 
const size_t QueNum = 200;

const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}


// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen; 
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;



const size_t ResVecLen = QueNum * RefPadBloLen;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

//cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);
double NikosTime= NikosTimer.elapsed();
std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/

/******************************************************************************************************************/
/* REFERENCE IMPLEMENTATION */
/* KERNEL NAME: PaPaRa-OpenCL-Reference.cl */

/* *
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter_Dual(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,InterRefVec_a_aux,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";


// Arguments
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

cl_mem InterRefVecPin_a_aux = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a_aux = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin_a_aux, myCommandQueue, InterRefVec_a_aux);
cl_mem InterRefVecDev_a_aux = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin_a_aux, InterRefVecDev_a_aux, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem InterIntVecDev = allocateDeviceMemory_32 (InterIntVecLen, myContext);
cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);


// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &InterRefVecDev_a_aux);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;
timer NikosTimerN;
clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);
std::cout << "\nGPU Time: " << NikosTimerN.elapsed() <<"\n\n";

// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

/* */

/******************************************************************************************************************/



/******************************************************************************************************************/
/* OPTIMIZED IMPLEMENTATION */
/* KERNEL NAME: PaPaRa-OpenCL-Optimized0.cl */

/* *
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter_Dual(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,InterRefVec_a_aux,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";


// Arguments
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

cl_mem InterRefVecPin_a_aux = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a_aux = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin_a_aux, myCommandQueue, InterRefVec_a_aux);
cl_mem InterRefVecDev_a_aux = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin_a_aux, InterRefVecDev_a_aux, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem InterIntVecDev = allocateDeviceMemory_32 (InterIntVecLen, myContext);
cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);


// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &InterRefVecDev_a_aux);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;
timer NikosTimerN;
clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);
std::cout << "\nGPU Time: " << NikosTimerN.elapsed() <<"\n\n";

// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

/* */

/******************************************************************************************************************/


/******************************************************************************************************************/
/* OPTIMIZED IMPLEMENTATION */
/* KERNEL NAME: PaPaRa-OpenCL-Optimized1.cl */

/* *
timer NikosTimerN;
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";


// Arguments
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

//cl_mem InterRefVecPin_a_aux = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
//InterRefVecPin_a_aux = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin_a_aux, myCommandQueue, InterRefVec_a_aux);
//cl_mem InterRefVecDev_a_aux = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
//copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin_a_aux, InterRefVecDev_a_aux, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem InterIntVecDev = allocateDeviceMemory_32 (InterIntVecLen, myContext);
cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);


// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
//clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &InterRefVecDev_a_aux);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);


// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "\nGPU Time: " << NikosTimerN.elapsed() <<"\n\n";

/* */

/******************************************************************************************************************/



/******************************************************************************************************************/
/* OPTIMIZED IMPLEMENTATION */
/* KERNEL NAME: PaPaRa-OpenCL-Optimized2.cl */

/* *
timer NikosTimerN;
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";

// Shared Memory Optimization
const size_t SharedLineSize = 4; // KBytes
const size_t SharedLineLength = (SharedLineSize * 1024)/4;
std::cout<<"Shared Memory Size:\t\t"<<SharedLineLength<<"\n";
const size_t SitesPerBlock = (SharedLineLength / RefPadBloLen) - 1 ;
std::cout<<"Sites Per Block:\t\t"<<SitesPerBlock<<"\n";
const size_t BlockIterations = InterRefVecLen / SharedLineLength + 1;
std::cout<<"Block Iterations:\t\t"<<BlockIterations<<"\n";
const size_t TotalSitesS = BlockIterations * SitesPerBlock;
std::cout<<"Total Sites:\t\t\t"<<TotalSitesS<<"\n";
const size_t BlockIterationsS = BlockIterations - 1;
const size_t GBlockSites = RefLen - BlockIterationsS * SitesPerBlock;
std::cout<<"Last Block Size:\t\t"<<GBlockSites<<"\n\n\n";

// Testing
//TODO: blocked on global with 2,3,4 blocks..

const size_t PerQueryNumberofConnectingEntriesGlobal = QueLenMax * RefPadBloLen;
const size_t TotalNumberofConnectingEntriesGlobal =  PerQueryNumberofConnectingEntriesGlobal * QueNum;
std::cout<<"Connections per Query:\t\t"<<PerQueryNumberofConnectingEntriesGlobal<<"\n";
std::cout<<"Connections TOTAL:\t\t"<<TotalNumberofConnectingEntriesGlobal<<"\n";

const size_t LastSLSCSDDevLen = TotalNumberofConnectingEntriesGlobal;




// Arguments
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;

//ArgVec[6] = BlockIterations-1;
//ArgVec[7] = SitesPerBlock;
//ArgVec[8] = GBlockSites;

for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;


// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem InterIntVecDev = allocateDeviceMemory_32 (InterIntVecLen, myContext);
cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);
cl_mem LastSLDev = allocateDeviceMemory_32 (LastSLSCSDDevLen, myContext);
cl_mem LastSCDev = allocateDeviceMemory_32 (LastSLSCSDDevLen, myContext);
cl_mem LastSDDev = allocateDeviceMemory_32 (LastSLSCSDDevLen+1, myContext);




// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
//clSetKernelArg(myKernel, 5, sizeof(int) * SharedLineLength, 0);
//clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &LastSLDev);
//clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &LastSCDev);
//clSetKernelArg(myKernel, 8, sizeof(cl_mem), (void *) &LastSDDev);
clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);

// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "\nGPU Time: " << NikosTimerN.elapsed() <<"\n\n";
/* */

/******************************************************************************************************************/



/******************************************************************************************************************/
/* OPTIMIZED IMPLEMENTATION */
/* KERNEL NAME: PaPaRa-OpenCL-Optimized3.cl */

/* *
timer NikosTimerN;
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";




// Blocked Version Parameters
const size_t NumberOfConnectingValuesPerVector = QueLenMax * RefPadBloLen * QueNum;
std::cout << "Number Of Connecting Values Per Vector:\t\t"<<NumberOfConnectingValuesPerVector<<"\n\n";

const size_t NumberOfSitesInBlockLine = 1500;
const size_t BlockLineLen = NumberOfSitesInBlockLine * RefPadBloLen;
const size_t GlobalBlockLineLen = BlockLineLen * QueNum;
std::cout << "Global Blocked Line Len:\t\t"<<GlobalBlockLineLen<<"\n\n";


// Arguments
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

//cl_mem InterRefVecPin_a_aux = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
//InterRefVecPin_a_aux = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin_a_aux, myCommandQueue, InterRefVec_a_aux);
//cl_mem InterRefVecDev_a_aux = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
//copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin_a_aux, InterRefVecDev_a_aux, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem InterIntVecDev = allocateDeviceMemory_32 (InterIntVecLen, myContext);
cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);
cl_mem SLVecDev = allocateDeviceMemory_32 (NumberOfConnectingValuesPerVector, myContext);
cl_mem SCVecDev = allocateDeviceMemory_32 (NumberOfConnectingValuesPerVector, myContext);
cl_mem BlockLineDev = allocateDeviceMemory_32 (GlobalBlockLineLen, myContext);



// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
//clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &InterRefVecDev_a_aux);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &SLVecDev);
clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &SCVecDev);
clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &BlockLineDev); 

clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);


// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "\nGPU Time: " << NikosTimerN.elapsed() <<"\n\n";

/* */

/******************************************************************************************************************/

/******************************************************************************************************************/
/* OPTIMIZED IMPLEMENTATION */
/* KERNEL NAME: PaPaRa-OpenCL-Optimized4.cl */

/* *
timer NikosTimerN;
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";




// Blocked Version Parameters
const size_t NumberOfConnectingValuesPerVector = QueLenMax * RefPadBloLen * QueNum;
std::cout << "Number Of Connecting Values Per Vector:\t\t"<<NumberOfConnectingValuesPerVector<<"\n\n";

const size_t NumberOfSitesInBlockLine = 1500;
const size_t BlockLineLen = NumberOfSitesInBlockLine * RefPadBloLen;
const size_t GlobalBlockLineLen = BlockLineLen * QueNum;
std::cout << "Global Blocked Line Len:\t\t"<<GlobalBlockLineLen<<"\n\n";


// Arguments
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

//cl_mem InterRefVecPin_a_aux = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
//InterRefVecPin_a_aux = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin_a_aux, myCommandQueue, InterRefVec_a_aux);
//cl_mem InterRefVecDev_a_aux = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
//copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin_a_aux, InterRefVecDev_a_aux, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem InterIntVecDev = allocateDeviceMemory_32 (InterIntVecLen, myContext);
cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);
cl_mem SLVecDev = allocateDeviceMemory_32 (NumberOfConnectingValuesPerVector, myContext);
cl_mem SCVecDev = allocateDeviceMemory_32 (NumberOfConnectingValuesPerVector, myContext);
cl_mem BlockLineDev = allocateDeviceMemory_32 (GlobalBlockLineLen, myContext);



// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
//clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &InterRefVecDev_a_aux);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &SLVecDev);
clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &SCVecDev);
clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &BlockLineDev); 

clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);


// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "\nGPU Time: " << NikosTimerN.elapsed() <<"\n\n";

/* */

/******************************************************************************************************************/








/******************************************************************************************************************/
/* OPTIMIZED IMPLEMENTATION */
/* KERNEL NAME: PaPaRa-OpenCL-Optimized5.cl */

/* *
timer NikosTimerN;
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";


// Blocked Version Parameters
// Gblock is Global Block
// Sblock is Shared Block

const size_t SitesInGblock = QueLenMax + 150;
const size_t SitesInSblock = RefLen - SitesInGblock ;

const size_t GblockLine = SitesInGblock * RefPadBloLen;
const size_t SblockLine = SitesInSblock * RefPadBloLen;

const size_t GblockLineLen = GblockLine * QueNum;
const size_t SblockLineLen = SblockLine * QueNum;

const size_t SLSCVecLen = QueLenMax * RefPadBloLen * QueNum;





// Arguments
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

//cl_mem InterRefVecPin_a_aux = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
//InterRefVecPin_a_aux = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin_a_aux, myCommandQueue, InterRefVec_a_aux);
//cl_mem InterRefVecDev_a_aux = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
//copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin_a_aux, InterRefVecDev_a_aux, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem InterIntVecDev = allocateDeviceMemory_32 (InterIntVecLen, myContext);
cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);

cl_mem GBlockLineDev1 = allocateDeviceMemory_32 (GblockLineLen, myContext);
cl_mem SBlockLineDev  = allocateDeviceMemory_32 (SblockLineLen, myContext);

cl_mem SLVecDev = allocateDeviceMemory_32 (SLSCVecLen , myContext);
cl_mem SCVecDev = allocateDeviceMemory_32 (SLSCVecLen , myContext);

// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
//clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &InterRefVecDev_a_aux);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);

clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &GBlockLineDev1);
clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &SBlockLineDev);

clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &SLVecDev);
clSetKernelArg(myKernel, 8, sizeof(cl_mem), (void *) &SCVecDev);

clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);


// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "\nGPU Time: " << NikosTimerN.elapsed() <<"\n\n";

/* */

/******************************************************************************************************************/


/******************************************************************************************************************/
/* OPTIMIZED IMPLEMENTATION */
/* KERNEL NAME: PaPaRa-OpenCL-Optimized6.cl */

/* *
timer NikosTimerN;
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";

// Blocked Version Parameters
const size_t BLOCKlengthCOLUMNS = 30;

const size_t tempcolGlen  = QueNum * QueLenMax * RefPadBloLen;
const size_t templineGlen = QueNum * (QueLenMax + BLOCKlengthCOLUMNS) * RefPadBloLen;
const size_t templineSlen = QueNum * BLOCKlengthCOLUMNS * RefPadBloLen;



// Arguments
const size_t ArgVecLen = QueNum + 7; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
ArgVec[6] = BLOCKlengthCOLUMNS;

for(size_t i = 7; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-7].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

//cl_mem InterRefVecPin_a_aux = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
//InterRefVecPin_a_aux = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin_a_aux, myCommandQueue, InterRefVec_a_aux);
//cl_mem InterRefVecDev_a_aux = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
//copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin_a_aux, InterRefVecDev_a_aux, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem InterIntVecDev = allocateDeviceMemory_32 (InterIntVecLen, myContext);
cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);


cl_mem SLVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
cl_mem SCVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
cl_mem BLOCKAW_lineDev = allocateDeviceMemory_32 (templineGlen , myContext);
cl_mem BLOCK_lineDev = allocateDeviceMemory_32 (templineSlen, myContext);


// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
//clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &InterRefVecDev_a_aux);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);

clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &SLVecDev);
clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &SCVecDev);
clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &BLOCKAW_lineDev);
clSetKernelArg(myKernel, 8, sizeof(cl_mem), (void *) &BLOCK_lineDev);

clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);


// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "\nGPU Time:\t\t" << NikosTimerN.elapsed() <<"\n\n";

/* */

/******************************************************************************************************************/



/******************************************************************************************************************/
/* OPTIMIZED IMPLEMENTATION */
/* KERNEL NAME: PaPaRa-OpenCL-Optimized7.cl - Optimized8 STANDARD GENERIC - UNFOLD8*/

/* *
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";


// Shared Memory 
const size_t shared_memory_size = 4;//KBs
const size_t templineSMaxLen = (shared_memory_size * 1024)/4; // integers




// Blocked Version Parameters
const size_t BLOCKlengthCOLUMNS = templineSMaxLen / RefPadBloLen;
std::cout<<"Block size in columns:\t\t"<<BLOCKlengthCOLUMNS<<"\n";
const size_t tempcolGlen  = QueNum * QueLenMax * RefPadBloLen;
const size_t templineGlen = QueNum * (QueLenMax + BLOCKlengthCOLUMNS) * RefPadBloLen;
const size_t templineSlen = QueNum * BLOCKlengthCOLUMNS * RefPadBloLen;




int total = RefLen;
int partF = QueLenMax; 
int partL = QueLenMax;
int partM = total - partF - partL;
int blockiterations = partM / BLOCKlengthCOLUMNS;
int sitesinblocks = blockiterations * BLOCKlengthCOLUMNS;
int partLext = partM - sitesinblocks;
    partL = partL + partLext;
int lastblock_refstart = total - partL;






// Arguments
const size_t ArgVecLen = QueNum + 8; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = QueLenMax;
ArgVec[5] = BLOCKlengthCOLUMNS;
ArgVec[6] = blockiterations;
ArgVec[7] = lastblock_refstart;

for(size_t i = 8; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-8].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);

//cl_mem SLVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
//cl_mem SCVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
cl_mem SLSCVecDev = allocateDeviceMemory_u32 (tempcolGlen, myContext);
cl_mem BLOCKAW_lineDev = allocateDeviceMemory_32 (templineGlen , myContext);
//cl_mem BLOCK_lineDev = allocateDeviceMemory_32 (templineSlen, myContext);


// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &ResVecDev);

//clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &SLVecDev);
//clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &SCVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &BLOCKAW_lineDev);
clSetKernelArg(myKernel, 5, sizeof(int)*templineSMaxLen, 0);
clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &SLSCVecDev);
clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;
timer NikosTimerN;
clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);
std::cout << "\nGPU Time:\t\t" << NikosTimerN.elapsed() <<"\n\n";


// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);


/* */

/******************************************************************************************************************/



/******************************************************************************************************************/
/* COMPRESSED IMPLEMENTATION */
/* KERNEL NAME: COMPRESSED*/

/* */
timer NikosTimerN;
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";


// Shared Memory 
const size_t shared_memory_size = 12;//KBs
const size_t templineSMaxLen = (shared_memory_size * 1024)/4; // integers




// Blocked Version Parameters
const size_t BLOCKlengthCOLUMNS = templineSMaxLen / RefPadBloLen;
std::cout<<"Block size in columns:\t\t"<<BLOCKlengthCOLUMNS<<"\n";
const size_t tempcolGlen  = QueNum * QueLenMax * RefPadBloLen;
const size_t templineGlen = QueNum * (QueLenMax + BLOCKlengthCOLUMNS) * RefPadBloLen;
const size_t templineSlen = QueNum * BLOCKlengthCOLUMNS * RefPadBloLen;




int total = RefLen;
int partF = QueLenMax; 
int partL = QueLenMax;
int partM = total - partF - partL;
int blockiterations = partM / BLOCKlengthCOLUMNS;
std::cout<<"Iterations:\t\t"<<blockiterations<<"\n";
int sitesinblocks = blockiterations * BLOCKlengthCOLUMNS;
int partLext = partM - sitesinblocks;
    partL = partL + partLext;
int lastblock_refstart = total - partL;

std::cout<<"Limit1:"<< partF <<"-- Limit2:"<<lastblock_refstart<<"Check:"<<partF+blockiterations*BLOCKlengthCOLUMNS+partL;

// Inter-Reference Memory Middle Part Compression: 4 ref entries per int
const size_t compression_effort = 6;
const size_t compressed_vector_length = ((lastblock_refstart - partF)/compression_effort)*RefPadBloLen;
std::cout<<"Compressed Vector Length:"<<compressed_vector_length;
int * CompRefMidVec = (int *) malloc(sizeof(int)*compressed_vector_length);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterIntComp(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,partF, lastblock_refstart,CompRefMidVec,RefPadBloLen);}






// Arguments
const size_t ArgVecLen = QueNum + 8; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = QueLenMax;
ArgVec[5] = BLOCKlengthCOLUMNS;
ArgVec[6] = blockiterations;
ArgVec[7] = lastblock_refstart;

for(size_t i = 8; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-8].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

cl_mem InterRefVecCompPin_a = allocatePinnedHostMemory_32 (compressed_vector_length, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (compressed_vector_length, InterRefVecCompPin_a, myCommandQueue, CompRefMidVec);
cl_mem InterRefVecCompDev_a = allocateDeviceMemory_32 (compressed_vector_length, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (compressed_vector_length, InterRefVecCompPin_a, InterRefVecCompDev_a, myCommandQueue);

cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);

//cl_mem SLVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
//cl_mem SCVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
cl_mem SLSCVecDev = allocateDeviceMemory_u32 (tempcolGlen, myContext);
cl_mem BLOCKAW_lineDev = allocateDeviceMemory_32 (templineGlen , myContext);
//cl_mem BLOCK_lineDev = allocateDeviceMemory_32 (templineSlen, myContext);


// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &ResVecDev);

//clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &SLVecDev);
//clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &SCVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &BLOCKAW_lineDev);
clSetKernelArg(myKernel, 5, sizeof(int)*templineSMaxLen, 0);
clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &SLSCVecDev);
clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &InterRefVecCompDev_a);
clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);
std::cout << "\nGPU Time:\t\t" << NikosTimerN.elapsed() <<"\n\n";


// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);


/* */

/******************************************************************************************************************/






/* SIMON'S REFERENCE TEST */

unsigned int score_ref;
//for(size_t i=0;i<1;i++)
//{
//score_ref = align_pvec_score( m_ref_pvecs[1], m_ref_aux[1], m_qs_pvecs[198], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
//std::cout << "Simon score for 0 0 0     " << score_ref <<"\n";
//}
unsigned int * ResVecRef = (unsigned int *) malloc (sizeof(unsigned int)*ResVecLen);
size_t y = 0, yy=0;
unsigned int fillzeros = RefPadBloLen - RefNum;
unsigned int ncups = 0;
align_arrays arr;

timer SimonTimerN;
for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) 
{
 	for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) 
	{          
                unsigned int score = align_pvec_score( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		//ResVec[y] = align_pvec_scoreBLOCKED( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		ResVecRef[y++] = score;	
        }  
	for(size_t yy=0;yy<fillzeros;yy++)
	{
		//ResVec[y] = 0;
		ResVecRef[y++] = 0;
 	}          
}
std::cout << "\nCPU Time:\t\t" << SimonTimerN.elapsed() <<"\n\n";


unsigned int totalcounter=0;

//std::cout << "GNcupsCPU: " << ncups / (NikosTime * 1e9) << "\n";
//std::cout << "GNcupsCPU: " << ncups / (SimonTimerN.elapsed() * 1e9) << "\n";
for(size_t i=0;i<ResVecLen;i++)
{
	
	if(ResVec[i]!=ResVecRef[i])
	{
		std::cout<<i<<":"<<ResVec[i]<<" and " << ResVecRef[i]<<"\n";
		totalcounter++;
	}
}

std::cout<<"\n\nIt is:"<<totalcounter<<" out of:"<<ResVecLen<<"\n\n";


/*
y = 0, yy=0;
timer NIKOSTimerN;
for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) 
{
 	for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) 
	{          
                //unsigned int score = align_pvec_score( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		ResVec[y++] = align_pvec_scoreBLOCKED_OPTIMIZED_3( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		//ResVecRef[y++] = score;	
        }  
	for(size_t yy=0;yy<fillzeros;yy++)
	{
		ResVec[y++] = 0;
		//ResVecRef[y++] = 0;
 	}          
}
std::cout << "\nCPU Time NEW:\t\t" << NIKOSTimerN.elapsed() <<"\n\n";


totalcounter=0;

//std::cout << "GNcupsCPU: " << ncups / (NikosTime * 1e9) << "\n";
//std::cout << "GNcupsCPU: " << ncups / (SimonTimerN.elapsed() * 1e9) << "\n";
for(size_t i=0;i<ResVecLen;i++)
{
	
	if(ResVec[i]!=ResVecRef[i])
	{
		std::cout<<i<<":"<<ResVec[i]<<" and " << ResVecRef[i]<<"\n";
		totalcounter++;
	}
}

std::cout<<"\n\nIt is:"<<totalcounter<<" out of:"<<ResVecLen<<"\n\n";

/****************/



        
    }
};



int main() {
    testbench tb( "qs.bin", "ref.bin" );
    tb.run_seq();
}
