#include <vector>
#include <string>
#include <iostream>
#include <stdexcept>

#include <cstdio>
#include <cstring>
#include <cassert>

#include "CL/opencl.h"

/*********************************************************************************************/

cl_platform_id select_NVIDIA_platform() {
	//Get number of available platforms and select the NVIDIA platform
	cl_uint num_platforms;
	cl_int ciErrNum;
	//cl_platform_id* clPlatformIDs;
	cl_platform_id selected_cl_platform_id = 0;
	char chBuffer[1024];

	ciErrNum = clGetPlatformIDs(0, NULL, &num_platforms);
	std::vector<cl_platform_id> clPlatformIDs;

	if (ciErrNum == CL_SUCCESS) {
		//printf("Detected %d platforms\n",num_platforms);

//		clPlatformIDs = (cl_platform_id*) malloc(
//				num_platforms * sizeof(cl_platform_id));

		clPlatformIDs.resize( num_platforms );


		ciErrNum = clGetPlatformIDs(num_platforms, &clPlatformIDs[0], NULL);

		//printf("Available platforms:\n");

		for (unsigned int i = 0; i < num_platforms; ++i) {
			ciErrNum = clGetPlatformInfo(clPlatformIDs[i], CL_PLATFORM_NAME,
					1024, &chBuffer, NULL);

			if (ciErrNum == CL_SUCCESS) {
				//printf("platform %d: %s\n", i, chBuffer);
				if (strstr(chBuffer, "NVIDIA") != NULL)
				//		if(strstr(chBuffer, "Intel") != NULL)
				{
					//printf("selected platform %d\n", i);
					selected_cl_platform_id = clPlatformIDs[i];
					break;
				}
			}
		}

	}

	if (selected_cl_platform_id != 0) {
		clGetPlatformInfo(selected_cl_platform_id, CL_PLATFORM_NAME, 1024,
				&chBuffer, NULL);
		printf("\nSelected platform: %s\n", chBuffer);
	}

	return selected_cl_platform_id;

}
/*********************************************************************************************/

cl_device_id  select_GPU_device(cl_platform_id  selected_platform) {

	// Get number of GPUs available
	cl_uint numDevices;
	cl_int ciErrNum;
	cl_device_id  selected_device_ID = 0;
	char chBuffer[1024];

	ciErrNum = clGetDeviceIDs(selected_platform, CL_DEVICE_TYPE_ALL, 0, NULL,
			&numDevices);

	std::vector<cl_device_id> deviceIDlist;

	if (ciErrNum == CL_SUCCESS) {
		//printf("\nDetected %d devices\n",numDevices);

		//cl_device_id * deviceIDlist = new cl_device_id[numDevices];

		deviceIDlist.resize( numDevices );
		clGetDeviceIDs(selected_platform, CL_DEVICE_TYPE_ALL, numDevices,
				&deviceIDlist[0], NULL);

		for (unsigned int i = 0; i < numDevices; i++) {
			//printf("GPU device %d:",i);

			clGetDeviceInfo(deviceIDlist[i], CL_DEVICE_TYPE_ALL, 1024,
					&chBuffer, NULL);

			//printf("platform %d: %s\n", i, chBuffer);
			if (strstr(chBuffer, "NVIDIA") != NULL)
			//if(strstr(chBuffer, "Intel") != NULL)
			{
				//printf("selected platform %d\n", i);
				selected_device_ID = deviceIDlist[i];
				break;
			}

		}
	}

	if (selected_device_ID != 0) {
		clGetDeviceInfo(selected_device_ID, CL_DEVICE_TYPE_GPU, 1024,
				&chBuffer, NULL);
		printf("\nSelected device: %s\n", chBuffer);
	}

	return selected_device_ID;
}

/*********************************************************************************************/
void opencl_notify (const char *errinfo, const void *private_info,size_t cb, void *user_data ) {
	std::cerr << "opencl started crying: '" << errinfo << "'\n";

	throw std::runtime_error( "opencl_notify. throwing exception."); // I guess it is not a good idea normally.
}

cl_context create_context(cl_device_id selected_device) {

	//cl_device_id constSelectedDeviceID = selected_device;

	// Create context

	cl_context mycontext;
	cl_int ciErrNum;

	mycontext = clCreateContext(0, 1, &selected_device, &opencl_notify, 0,
			&ciErrNum);

	return mycontext;
}

/*********************************************************************************************/

cl_command_queue create_command_queue(cl_device_id selected_device,
		cl_context context) {
	cl_int ciErrNum;

	cl_command_queue mycmdqueue;

	mycmdqueue = clCreateCommandQueue(context, selected_device, 0, &ciErrNum);

	return mycmdqueue;

}

/*********************************************************************************************/

std::string load_kernel_code(const char * filename ) {

	FILE *fp = fopen(filename, "r");
	//char * cl_source;



	if (fp != NULL) {
		fseek(fp, 0, SEEK_END);
		size_t size = ftell(fp);
		fseek(fp, 0, SEEK_SET);

		//cl_source = (char *) malloc(sizeof(char) * (*size + 1));
		std::vector<char> data(size);

		size = fread(data.data(), 1, data.size(), fp);
		assert( size == data.size() );

		fclose(fp);


		//data.back() = '\0';

		//show_cl_source(cl_source, *size);

		return std::string( data.begin(), data.end() ); // TODO: change to move semantics

	} else {
		printf("\n%s is missing!\n", filename);
		return std::string();
	}

}

/*********************************************************************************************/

cl_program create_and_build_program(cl_device_id selected_device,
		cl_context mycontext, const char * kernel_code, int kernel_size) {

	cl_program myprogram;
	cl_int ciErrNum;
	size_t myprogram_length = (size_t) kernel_size;

	myprogram = clCreateProgramWithSource(mycontext, 1,
			(const char **) &kernel_code, &myprogram_length, &ciErrNum);

	// Build program and check build status

	clBuildProgram(myprogram, 0, NULL, NULL, NULL, NULL);

	cl_build_status build_status;

	clGetProgramBuildInfo(myprogram, selected_device, CL_PROGRAM_BUILD_STATUS,
			sizeof(cl_build_status), &build_status, NULL);

	if (build_status == CL_BUILD_SUCCESS) {
		printf("\n\n>>> GPU STATUS: ALL SET! <<<\n\n\n\n");
	} else {
		printf("\n\n>>> GPU STATUS: ISSUE CAME UP! <<<\n\n");
	}

	return myprogram;
}

/*********************************************************************************************/

cl_kernel create_kernel(cl_program myprogram, const char * kernel_name) {
	cl_kernel mykernel;
	cl_int ciErrNum;

	mykernel = clCreateKernel(myprogram, kernel_name, &ciErrNum);

	assert( ciErrNum == CL_SUCCESS);

	return mykernel;
}

