#include <iostream>
#include <fstream>
#include <stdexcept>
#include <vector>
#include <stdint.h>
#include <climits>
#include <algorithm>
#include <sys/time.h>

#include "CL/opencl.h"
//#include <shrQATest.h>

#include "setupGPU.h"
#include "memCNTRL.h"
#include "utilsPaPaRa.h"


static double gettime(void )
{
    struct timeval ttime;
    gettimeofday(&ttime , 0);
    return ttime.tv_sec + ttime.tv_usec * 0.000001;
}


class timer {
    double m_st;
    
public:
    timer() : m_st(gettime()) {};
    
    double elapsed() const {
        return gettime() - m_st;
    }
    
    
};



typedef int score_t;
typedef unsigned short score_t_exp;
typedef unsigned int u32_de;


struct align_arrays {
    std::vector<score_t> s;
   // std::vector<score_t> si;
};


static score_t align_pvec_score( std::vector<int> &a, std::vector<unsigned int> &a_aux, const std::vector<uint8_t> &b, score_t mismatch_score, score_t match_cgap, score_t gap_open, score_t gap_extend, align_arrays &arr ) {

    if( arr.s.size() < a.size()  ) {
        arr.s.resize( a.size() );
       // arr.si.resize( a.size() );
    }
    
    const score_t aux_cgap = 0x1;
    
    const size_t band_width = a.size() - b.size();
    std::fill( arr.s.begin(), arr.s.end(), 0 );
    
    const score_t LARGE = 32000;
//     std::fill( arr.si.begin(), arr.si.end(), LARGE );
    arr.s.at(band_width+1) = LARGE;

    score_t min_score = LARGE;
    
    for( size_t ib = 0; ib < b.size(); ib++ ) {
        int bc = b[ib];
        
        score_t last_sl = LARGE;
        score_t last_sc = LARGE;
        score_t last_sdiag = 0.0;
        
        size_t astart = ib;
                
        score_t * __restrict s_iter = &arr.s[0];

        
        bool lastrow = ib == (b.size() - 1);
        
        last_sdiag = *s_iter;
        
        
        for( size_t ia = astart; ia <= ib + band_width; ++ia, ++s_iter ) {  
            score_t ac = a[ia];
            const bool cgap = a_aux[ia] == aux_cgap;
            
            // determine match or mis-match according to parsimony bits coming from the tree.
            bool is_match = ( ac & bc ) != 0;

            score_t sm = last_sdiag;
            
            if( !is_match ) {
                sm += mismatch_score;
            }
            
            last_sdiag = *(s_iter+1);

            score_t last_sc_OPEN; 
            score_t sl_score_stay;
            
            if( cgap ) {
                last_sc_OPEN = last_sc; 
                sl_score_stay = last_sl;
                sm += match_cgap;
            } else {
                last_sc_OPEN = last_sc + gap_open + gap_extend;  
                sl_score_stay = last_sl + gap_open;
            }
            
            score_t sl = std::min( sl_score_stay, last_sc_OPEN );
                        
            last_sl = sl;
            

            score_t su = last_sdiag + mismatch_score;
            
//             std::cout << "x: " << sm << " " << su << " " << sl << "  ";
            score_t sc = std::min( sm, std::min( su, sl ) );
            
            last_sc = sc;
            *s_iter = sc;
            
        
            if( lastrow ) {
                if( sc < min_score ) {
                    
                    min_score = sc;
                }
            }
           
        }
//         std::cout << "\n";
//         getchar();
    }

    return min_score;
    
}



static score_t_exp align_pvec_score_experimental( std::vector<int> &a, std::vector<unsigned int> &a_aux, const std::vector<uint8_t> &b, score_t_exp mismatch_score, score_t_exp match_cgap, score_t_exp gap_open, score_t_exp gap_extend, align_arrays &arr ) {

	
    size_t pseudovecInd=0;

    size_t temparrInd=0;

    score_t_exp * temparray = (score_t_exp*) malloc(sizeof(score_t_exp)*a.size());

    /****TEST CODE****/ 
	
    size_t de_t_vec_sz = 0;
    size_t de_t_vec_ind = 0;
    
    de_t_vec_sz = a.size()%2==0?a.size()/2:(a.size()+1)/2; 

    u32_de * tmparr_dual = (u32_de*) malloc(sizeof(u32_de)*de_t_vec_sz);
    
    //std::cout << a.size() <<"  "<< de_t_vec_sz<<"\n";
    /*****************/
    



    const score_t_exp aux_cgap = 0x1; //de operation cost = 0;
    const u32_de aux_cgap_de = 65537;
      
    const size_t band_width = a.size() - b.size();
    const size_t band_width_de = de_t_vec_sz - b.size();


    for(size_t i=0;i<a.size();i++)
    {
	//temparray[i]=0.0;
    }
    
    const score_t_exp LARGE = 32000; //de operation cost = 0;
    const u32_de LARGE_DE = 2097184000;


    temparray[band_width+1] = LARGE; //de operation cost = 0;
    tmparr_dual[band_width_de+1] = LARGE;

    score_t_exp min_score = LARGE; //de operation cost = 0;
    u32_de min_score_de = LARGE_DE;

    for( size_t ib = 0; ib < b.size(); ib++ ) {
        u32_de bc = b[ib]; 
	
	//u32_de bc_de = b[ib];
	//bc_de = bc_de << 16;
	//bc_de = bc_de | b[ib+1];
        
        score_t_exp last_sl = LARGE; 
        score_t_exp last_sc = LARGE; 
        score_t_exp last_sdiag = 0.0; 

	//u32_de last_sl_de = LARGE_DE; 
       // u32_de last_sc_de = LARGE_DE; 
        u32_de last_sdiag_de = 0.0;
        
        size_t astart = ib;

        
        bool lastrow = ib == (b.size() - 1);
        temparrInd=0;
        last_sdiag = temparray[temparrInd];

	de_t_vec_ind=0;
	last_sdiag_de = tmparr_dual[de_t_vec_ind];


        
        
        for( size_t ia = astart; ia <= ib + band_width; ++ia ) { 

 	    /************************/
            score_t_exp ac = a[ia];

	    u32_de ac_de_1 = a[ia];
	    u32_de ac_de_2 = a[ia+1];

	    u32_de a_aux_de_1 = a_aux[ia];
	    u32_de a_aux_de_2 = a_aux[ia+1];

         
            bool cgap = a_aux[ia] == aux_cgap;

	    const bool cgap_de_1 = a_aux_de_1 == aux_cgap;
	    const bool cgap_de_2 = a_aux_de_2 == aux_cgap; 



            bool is_match = ( ac & bc ) != 0;

	    bool is_match_de_1 = ( ac_de_1 & bc ) != 0;
	    bool is_match_de_2 = ( ac_de_2 & bc ) != 0;
		
	    



 
	    
	    score_t_exp sm = last_sdiag;
            
            if( !is_match ) {
                sm += mismatch_score;
            }
            
	    temparrInd++;
            last_sdiag = temparray[temparrInd];

            score_t_exp last_sc_OPEN; 
            score_t_exp sl_score_stay;
            
            if( cgap ) {
                last_sc_OPEN = last_sc; 
                sl_score_stay = last_sl;
                sm += match_cgap;
            } else {
                last_sc_OPEN = last_sc + gap_open + gap_extend;  
                sl_score_stay = last_sl + gap_open;
            }
            
            //score_t_exp sl = std::min( sl_score_stay, last_sc_OPEN );
            unsigned int sl=sl_score_stay<last_sc_OPEN?sl_score_stay:last_sc_OPEN;           
            last_sl = sl;
            

            score_t_exp su = last_sdiag + mismatch_score;
            

            score_t_exp sc = su<sl?su:sl;
	    sc = sc<sm?sc:sm;
            
            last_sc = sc;
	    temparray[temparrInd-1] = sc;

        
            if( lastrow ) {
                if( sc < min_score ) {
                    
                    min_score = sc;
                }
            }
            /************************/
        
           
        }
    }

    return min_score;
    
}


static score_t_exp align_pvec_score_experimental3( unsigned int * referenceSequence, unsigned int * querySequence, unsigned int referenceLength, unsigned int queryLength, score_t_exp mismatch_score, score_t_exp match_cgap, score_t_exp gap_open, score_t_exp gap_extend, align_arrays &arr ) {

	
    size_t pseudovecInd=0;

    size_t temparrInd=0;

    unsigned int * temparray = (unsigned int*) malloc(sizeof(unsigned int)*referenceLength);


    



    const unsigned int aux_cgap = 0x1; //de operation cost = 0;

      
    const size_t band_width = referenceLength - queryLength;



    for(size_t i=0;i<referenceLength;i++)
    {
	temparray[i]=0.0;
    }
    
    const unsigned int LARGE = 32000; //de operation cost = 0;



    temparray[band_width+1] = LARGE; //de operation cost = 0;


    unsigned int min_score = LARGE; //de operation cost = 0;


    for( size_t ib = 0; ib < queryLength; ib++ ) {
        u32_de bc = querySequence[ib]; 	

        
        score_t_exp last_sl = LARGE; 
        score_t_exp last_sc = LARGE; 
        score_t_exp last_sdiag = 0.0; 

        
        size_t astart = ib;

        
        bool lastrow = ib == (queryLength - 1);

        temparrInd=0;
        last_sdiag = temparray[temparrInd];



        
        
        for( size_t ia = astart; ia <= ib + band_width; ++ia ) { 

 	
            unsigned int ac = referenceSequence[ia]>>1;

         
            bool cgap = (referenceSequence[ia]&1) == aux_cgap;




            bool is_match = ( ac & bc ) != 0;

		
	    



 
	    
	    unsigned int sm = last_sdiag;
            
            if( !is_match ) {
                sm += mismatch_score;
            }
            
	    temparrInd++;
            last_sdiag = temparray[temparrInd];

            unsigned int last_sc_OPEN; 
            unsigned int sl_score_stay;
            
            if( cgap ) {
                last_sc_OPEN = last_sc; 
                sl_score_stay = last_sl;
                sm += match_cgap;
            } else {
                last_sc_OPEN = last_sc + gap_open + gap_extend;  
                sl_score_stay = last_sl + gap_open;
            }
            
            //score_t_exp sl = std::min( sl_score_stay, last_sc_OPEN );
            unsigned int sl=sl_score_stay<last_sc_OPEN?sl_score_stay:last_sc_OPEN;           
            last_sl = sl;
            

            unsigned int su = last_sdiag + mismatch_score;
            

            unsigned int sc = su<sl?su:sl;
	    sc = sc<sm?sc:sm;
            
            last_sc = sc;
	    temparray[temparrInd-1] = sc;

        
            if( lastrow ) {
                if( sc < min_score ) {
                    
                    min_score = sc;
                }
            }
   
        
           
        }
    }

    return min_score;
    
}
         /************************/



class testbench {
    std::vector<std::vector <int> > m_ref_pvecs;
    std::vector<std::vector <unsigned int> > m_ref_aux;
    
    std::vector<std::vector <uint8_t> > m_qs_pvecs;

    
    void load_qs( std::istream &is ) {
        size_t num_qs;
        is >> num_qs;
        
        m_qs_pvecs.resize( num_qs );
        
        for( int i = 0; i < m_qs_pvecs.size(); ++i ) {
            size_t len;
            is >> len;
            while( isspace( is.get() ) && !is.eof() ) {}
            is.unget();
            
            if( is.eof() ) {
                throw std::runtime_error( "unexpected end of file" );
            }
            
            m_qs_pvecs[i].resize(len);
            is.read( (char*)&m_qs_pvecs[i][0], len );
//             std::cout << "qs: " << i << " " << m_qs_pvecs[i].size() << "\n";
        }
    }
    
    void load_ref( std::ifstream &is ) {
        size_t num_ref;
        is >> num_ref;
        m_ref_pvecs.resize( num_ref );
        m_ref_aux.resize( num_ref );
        
        for( int i = 0; i < m_ref_pvecs.size(); ++i ) {
            size_t len;
            is >> len;
            while( isspace( is.get() ) && !is.eof() ) {}
            is.unget();
            
            if( is.eof() ) {
                throw std::runtime_error( "unexpected end of file" );
            }
            
            m_ref_pvecs[i].resize(len);
            is.read( (char*)&m_ref_pvecs[i][0], len * sizeof(int) );
            
            m_ref_aux[i].resize(len);
            is.read( (char*)&m_ref_aux[i][0], len * sizeof(unsigned int) );
            
//             std::cout << "ref: " << i << " " << m_ref_pvecs[i].size() << "\n";
        }
        
    }
    template<typename T, size_t min_len>
    static void trim_ref( std::vector<T> &v ) {
        
    	const size_t max_len = min_len;
        while( v.size() < min_len ) {
            std::vector<T> v_new = v;
            
            
            std::copy( v.begin(), v.end(), std::back_inserter(v_new) );
            
            v.swap(v_new);
            
        }
        

        if( v.size() > max_len ) {
            v.resize( max_len );
        }
        
        
    }   
 
public:  
    testbench( const char *qs_name, const char *ref_name ) {
        {
            std::ifstream is( qs_name );
            load_qs( is );
        }
        {
            std::ifstream is( ref_name );
            load_ref( is );
        }

	std::srand(1234);
	std::random_shuffle( m_ref_pvecs.begin(), m_ref_pvecs.end() );
	std::srand(1234);
	std::random_shuffle( m_ref_aux.begin(), m_ref_aux.end() );
        
        
        // reduce size of testset
        size_t num_ref = 32; // JUMP
        size_t num_qs = 200;
        
        num_ref = std::min( num_ref, m_ref_pvecs.size() );
        num_qs = std::min( num_qs, m_qs_pvecs.size() );
        
        m_ref_pvecs.resize(num_ref);
        m_ref_aux.resize(num_ref);
        m_qs_pvecs.resize(num_qs);
        
	//const size_t len = 1276;
	//std::for_each( m_ref_pvecs.begin(), m_ref_pvecs.end(), &trim_ref<int,len> );
        //std::for_each( m_ref_aux.begin(), m_ref_aux.end(), &trim_ref<unsigned int,len> );

    }
    
    
    void run_seq() {
        const int score_gap_open = 1;
        const int score_gap_extend = 1;
        const int score_mismatch = 1;
        const int score_match_cgap = 4;
        
        
        
        uint64_t ncup = 0;
        
        
        /******************************************/
	/* Acceleration attempt implemented by me */
        /******************************************/

        /* Setting up GPU.. */

	int i;
	
	cl_int ciErrNum;

   	cl_platform_id* myPlatform = select_NVIDIA_platform();

    	cl_device_id * myDevice = select_GPU_device(myPlatform);

    	cl_context myContext= create_context(myDevice);

    	cl_command_queue myCommandQueue = create_command_queue (myDevice, myContext);

    	int kernel_size;

    	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_GLOBAL_MEM.cl", &kernel_size);
        //char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF.cl", &kernel_size);
	char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF2.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_SHARED_MEM.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_SHARED_MEM2.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_BLOCKED.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_BLOCKED3.cl", &kernel_size);
	//char * kernel_code = load_kernel_code ("GPU_PaPaRa_kernel_INTER_REF_BLOCKED_TEST.cl", &kernel_size);

    	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);

    	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_GLOBAL_MEM");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF");
	cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF2");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_SHARED_MEM");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_SHARED_MEM2");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_BLOCKED");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_BLOCKED3");
	//cl_kernel myKernel = create_kernel(myProgram, "GPU_PaPaRa_kernel_INTER_REF_BLOCKED_TEST");

        /* Memory reorganization 

	const size_t reflen = m_ref_pvecs[m_ref_pvecs.size()-1].size();

	//printf("\n\nReference length (characters): %d\n",reflen);

	const size_t compactRefMode = 1; // # of Character+cgap sets in u32
        const size_t alignBlockSize = 32;
	const size_t compactRefSize = getCompactRefSize (compactRefMode, reflen, alignBlockSize);
	
	//printf("\n\nCompact Size per Ref (u32): %d\n",compactRefSize);

	unsigned int * compactRef = getCompactReference (compactRefMode, reflen, m_ref_pvecs[0], m_ref_aux[0], compactRefSize);
	

	// Simon reference test

	timer SimonTimer;	
	int score_s;
	ncup=0; 
	for(size_t i=0;i<1;i++)
	{
	score_s = align_pvec_score( m_ref_pvecs[0], m_ref_aux[0], m_qs_pvecs[0], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
	 ncup += reflen * m_qs_pvecs[1].size();
	}
	//std::cout << "\n\nSimon score for 0 0 0     " << score_s << " in "<<SimonTimer.elapsed()<<"\n";
	//std::cerr << "GNcups: " << ncup / (SimonTimer.elapsed() * 1e9) << "\n";

 	
	// My experiment




	// Find Maximum Query Length
	//unsigned int maxQuerySize = 0; 	for(size_t i=0;i<m_qs_pvecs.size();i++) {if(maxQuerySize<m_qs_pvecs[i].size()){maxQuerySize=m_qs_pvecs[i].size();}}



/*******************************************
// Load Check
timer GPUTimer;
const size_t num_of_queries = m_qs_pvecs.size();
const size_t num_of_references = m_ref_pvecs.size();

const size_t reference_size = m_ref_pvecs[0].size();
unsigned int maxQuerySize = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(maxQuerySize<m_qs_pvecs[i].size()){maxQuerySize=m_qs_pvecs[i].size();}}
const size_t query_size_max = maxQuerySize;

//std::cout<< num_of_queries << " queries to be aligned to " << num_of_references << " references\n";
//std::cout<< query_size_max << " query length to be aligned to " << reference_size << " reference length\n";


const size_t memBlockSize = 32;
const unsigned int constQueryLength = get_constSequenceLength (query_size_max,memBlockSize,1);
const unsigned int constReferenceLength = get_constSequenceLength (reference_size,memBlockSize,1);



unsigned int referencesAllSize = constReferenceLength * num_of_references;
unsigned int * referencesAll = (unsigned int *) malloc(sizeof(unsigned int)*referencesAllSize);
for(size_t i=0;i<num_of_references;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],constReferenceLength,referencesAll);}

unsigned int queriesAllSize = constQueryLength * num_of_queries;
unsigned int * queriesAll = (unsigned int *) malloc(sizeof(unsigned int)*queriesAllSize);
for(size_t i=0;i<num_of_queries;i++){putSequenceQue(i,m_qs_pvecs[i],constQueryLength,queriesAll);}


cl_mem referencesAllPinned = allocatePinnedHostMemory_u32 (referencesAllSize, myContext, myCommandQueue);
referencesAllPinned = initializePinnedHostMemory_u32_Sequence (referencesAllSize, referencesAllPinned, myCommandQueue, referencesAll);
cl_mem referencesAllDevice = allocateDeviceMemory_u32 (referencesAllSize, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (referencesAllSize, referencesAllPinned, referencesAllDevice, myCommandQueue);
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &referencesAllDevice);

cl_mem queriesAllPinned = allocatePinnedHostMemory_u32 (queriesAllSize, myContext, myCommandQueue);
queriesAllPinned = initializePinnedHostMemory_u32_Sequence (queriesAllSize, queriesAllPinned, myCommandQueue, queriesAll);
cl_mem queriesAllDevice = allocateDeviceMemory_u32 (queriesAllSize, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (queriesAllSize, queriesAllPinned, queriesAllDevice, myCommandQueue);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &queriesAllDevice);




//std::cout<< constReferenceLength << " FIXED query length to be aligned to " << referencesAllSize << " reference length ALL\n";



//size_t testvalue = shrRoundUp((int)memBlockSize, 45);
//std::cout<<"the test value is:"<<testvalue<<"\n";

//FILE * fp;
//fp = fopen("test.txt","w");

//for(size_t i=0;i<queriesAllSize;i++)
//{
//	fprintf(fp,"%u-%u:%u %u\n",i,queriesAll[i], m_qs_pvecs[1][i],m_qs_pvecs[1].size());
//}

//fclose(fp);



//float sharedMemLineSize = get_shareMemLineSize (constReferenceLength);
//std::cout<< sharedMemLineSize << " Shared Memory Required \n";

/*******************************************/


/*** DUMB IMPLEMENTATION FOR GPU TESTING***

//for(size_t i=0;i<20;i++)
//{

const size_t argumentsSize = 32;
const size_t argmult = 1;
const size_t compReferenceMode = 1;
const size_t compQueryMode = 1;
const size_t alignMemBlockSize = 32;

const size_t test1 = 2;
const size_t test2 = 24;


const size_t referenceLength = m_ref_pvecs[m_ref_pvecs.size()-1].size();
const size_t compReferenceLength = getCompactSize (compReferenceMode, referenceLength, alignMemBlockSize);
const size_t queryLength = m_qs_pvecs[0].size();
//std::cout<<"Testing Reference Length:"<<compReferenceLength<<" and Query Length: "<< queryLength<<"\n\n";
	
unsigned int * compactReference = getCompactReference (compReferenceMode, referenceLength, m_ref_pvecs[0], m_ref_aux[0], compReferenceLength);
unsigned int * Query = getCompactQuery(compQueryMode, queryLength, m_qs_pvecs[0], queryLength);


cl_mem referencePinned = allocatePinnedHostMemory_u32 (compReferenceLength, myContext, myCommandQueue);
cl_mem queryPinned = allocatePinnedHostMemory_u32 (queryLength, myContext, myCommandQueue);
cl_mem argumentsPinned = allocatePinnedHostMemory_u32 (test1*test2, myContext, myCommandQueue);


referencePinned = initializePinnedHostMemory_u32_Sequence (compReferenceLength, referencePinned, myCommandQueue, compactReference);						
queryPinned = initializePinnedHostMemory_u32_Sequence (queryLength, queryPinned, myCommandQueue,Query);
argumentsPinned = initializePinnedHostMemory_u32_specialArguments (test1*test2, argumentsPinned, myCommandQueue,argumentsSize*argmult);

cl_mem referenceDevice = allocateDeviceMemory_u32 (compReferenceLength, myContext);
cl_mem queryDevice = allocateDeviceMemory_u32 (queryLength, myContext);
cl_mem argumentsDevice = allocateDeviceMemory_u32 (test1*test2, myContext);
cl_mem tempLineDevice = allocateDeviceMemory_u32 (referencesAllSize, myContext);


copyPinnedHostMemoryToDeviceMemory_u32 (compReferenceLength, referencePinned, referenceDevice, myCommandQueue);
copyPinnedHostMemoryToDeviceMemory_u32 (queryLength, queryPinned, queryDevice, myCommandQueue);
copyPinnedHostMemoryToDeviceMemory_u32 (test1*test2, argumentsPinned, argumentsDevice, myCommandQueue);

//clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &referenceDevice);
//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &queryDevice);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &argumentsDevice);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &tempLineDevice);

//clSetKernelArg(myKernel, 3, sizeof(unsigned int) * constReferenceLength, 0);

clFinish(myCommandQueue);

//size_t workGroupSize[2]; 
 
size_t workGroupSize[] = {test1*test2};
size_t workGroupSize2[] = {test2};//argumentsSize};
//cl_event event;
//clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, workGroupSize, workGroupSize2, 0, NULL, &event);

//clFinish(myCommandQueue);

unsigned int * sequenceReturned = copyDeviceMemoryToPinnedHostMemory_u32 (test1*test2, argumentsDevice, argumentsPinned, myCommandQueue);

clFinish(myCommandQueue);

//for(size_t yy=0; yy < test1*test2; yy++)
//{
//	printf("%d:%u\n",yy,sequenceReturned[yy]);	

//}
//std::cout << "\n\nGPU score for 0 0 0    in "<<GPUTimer.elapsed()<<"\n";


size_t workgroupsize;
clGetKernelWorkGroupInfo (myKernel, *myDevice, CL_KERNEL_WORK_GROUP_SIZE, sizeof(size_t), &workgroupsize,NULL);
//std::cout<<"Work Group Size:"<<workgroupsize<<"\n";

cl_ulong localmemsize;
clGetKernelWorkGroupInfo (myKernel, *myDevice, CL_KERNEL_LOCAL_MEM_SIZE, sizeof(cl_ulong), &localmemsize,NULL);
//std::cout<<"Local Mem Size:"<<localmemsize<<"\n";

/***************/

/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION: NO OPTIMIZATIONS 

timer NikosTimer;
const size_t RefNum = 2;
const size_t QueNum = 200;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 2; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing
ArgVec[0] = RefLen;
ArgVec[1] = QueNum;
for(size_t i = 2; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-2].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t ResVecLen = QueNum * RefNum;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &RefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &IntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeLocal[] = {RefNum};//argumentsSize};

cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

//for(size_t i=0;i<ResVecLen;i++)
//{
//	std::cout<<i<<":"<<ResVec[i]<<"\n";
//}
std::cout << "GPU Time: " << NikosTimer.elapsed();

/**/

/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-REFERENCE GLOBAL 

timer NikosTimer;
const size_t RefNum = 192; 
const size_t QueNum = 200;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = RefNum;//get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen;
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}

// SIZE CHECK AND FIT TO SHARED AND GLOBAL MEMORY // TODO IN NEXT STEP
//const size_t SharedMemSize = 48; // KBytes
//const size_t gMemSAddr = get_sGloIndex (SharedMemSize, RefPadBloLen*3);
//const size_t SharedMemSizePerMP = gMemSAddr;
//const size_t GlobalMemSizePerMP = InterRefVecLen - SharedMemSizePerMP;
//const size_t TotalGlobalMemSize = QueNum * GlobalMemSizePerMP;

//std::cout << "Total/ Global/ Shared"<< InterRefVecLen <<"/"<< GlobalMemSizePerMP <<"/" << SharedMemSizePerMP <<"\n";


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
//ArgVec[0] = RefLen;
//ArgVec[1] = QueNum;
//ArgVec[2] = RefPadBloLen;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;

//std::cout<<"Before/After:"<<IntVecLen<<"/"<<InterIntVecLen*4/1024/1024<<"\n";

const size_t ResVecLen = QueNum * RefPadBloLen;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &RefVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
//clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &IntVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);
double NikosTime= NikosTimer.elapsed();
std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/


/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-REFERENCE GLOBAL  2 -- WORKING DONE KERNEL NAME: INTER REF2

timer NikosTimer;
const size_t RefNum = 102;  
const size_t QueNum = 500;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = RefNum;//get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen;
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}

// SIZE CHECK AND FIT TO SHARED AND GLOBAL MEMORY // TODO IN NEXT STEP
const size_t SharedMemSize32 = 48; // KBytes
//const size_t gMemSAddr = get_sGloIndex (SharedMemSize, RefPadBloLen*3);
//const size_t SharedMemSizePerMP = gMemSAddr;
//const size_t GlobalMemSizePerMP = InterRefVecLen - SharedMemSizePerMP;
//const size_t TotalGlobalMemSize = QueNum * GlobalMemSizePerMP;

//std::cout << "Total/ Global/ Shared"<< InterRefVecLen <<"/"<< GlobalMemSizePerMP <<"/" << SharedMemSizePerMP <<"\n";


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
//ArgVec[0] = RefLen;
//ArgVec[1] = QueNum;
//ArgVec[2] = RefPadBloLen;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;

//std::cout<<"Before/After:"<<IntVecLen<<"/"<<InterIntVecLen*4/1024/1024<<"\n";

const size_t ResVecLen = QueNum * RefPadBloLen;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &RefVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
//clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &IntVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);
double NikosTime= NikosTimer.elapsed();
std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/


/******************************************************************************************************************************/
// ** WORKING IMPLEMENTATION:  INTER-QUERY GLOBAL -- KERNEL NAME: INTER QUE 

timer NikosTimer;
const size_t RefNum = 32; // JUMP 
const size_t QueNum = 200;

const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}


// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen; 
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;



const size_t ResVecLen = QueNum * RefPadBloLen;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

//cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);
double NikosTime= NikosTimer.elapsed();
std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/




/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-REFERENCE GLOBAL + SHARED MEMORY 

timer NikosTimer;
const size_t RefNum = 128;
const size_t QueNum = 200;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen;
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}

// SIZE CHECK AND FIT TO SHARED AND GLOBAL MEMORY // TODO IN NEXT STEP
const size_t SharedMemSize = 4; // KBytes
const size_t SharedMemSize32 = (SharedMemSize * 1024)/4;
const size_t gMemSAddr = get_sGloIndex (SharedMemSize, RefPadBloLen);

//const size_t SharedMemSizePerMP = gMemSAddr;
//const size_t GlobalMemSizePerMP = InterRefVecLen - SharedMemSizePerMP;
//const size_t TotalGlobalMemSize = QueNum * GlobalMemSizePerMP;

//std::cout << "Total/ Global/ Shared"<< InterRefVecLen <<"/"<< GlobalMemSizePerMP <<"/" << SharedMemSizePerMP <<"\n";


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 7; 
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);


ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
ArgVec[6] = gMemSAddr;

for(size_t i = 7; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-7].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;

const size_t ResVecLen = QueNum * RefPadBloLen;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &RefVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
//clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &IntVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clSetKernelArg(myKernel, 5, sizeof(unsigned int) * SharedMemSize32 , 0);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/



/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-REFERENCE GLOBAL + QUERIES IN SHARED MEM

timer NikosTimer;
const size_t RefNum = 128;
const size_t QueNum = 200;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen;
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}

// SIZE CHECK AND FIT TO SHARED AND GLOBAL MEMORY // TODO IN NEXT STEP
//const size_t SharedMemSize = 48; // KBytes
//const size_t gMemSAddr = get_sGloIndex (SharedMemSize, RefPadBloLen*3);
//const size_t SharedMemSizePerMP = gMemSAddr;
//const size_t GlobalMemSizePerMP = InterRefVecLen - SharedMemSizePerMP;
//const size_t TotalGlobalMemSize = QueNum * GlobalMemSizePerMP;

//std::cout << "Total/ Global/ Shared"<< InterRefVecLen <<"/"<< GlobalMemSizePerMP <<"/" << SharedMemSizePerMP <<"\n";


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
//ArgVec[0] = RefLen;
//ArgVec[1] = QueNum;
//ArgVec[2] = RefPadBloLen;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;

//std::cout<<"Before/After:"<<IntVecLen<<"/"<<InterIntVecLen*4/1024/1024<<"\n";

const size_t ResVecLen = QueNum * RefPadBloLen;

const size_t QueSMemLen = get_next32mult(QueLenMax);

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &RefVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
//clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &IntVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clSetKernelArg(myKernel,5,sizeof(unsigned int)*QueSMemLen,0);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/



/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-REFERENCE GLOBAL - BLOCKED

timer NikosTimer;
const size_t RefNum =5; 
const size_t QueNum = 200;
 size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen;
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}

// SIZE CHECK AND FIT TO SHARED AND GLOBAL MEMORY // TODO IN NEXT STEP
const size_t SharedMemSize = 5; // KBytes
const size_t SharedMemSize32 = (SharedMemSize * 1024) / 4 ;
const size_t BlockSizeInColumns= (SharedMemSize32 / RefPadBloLen) + 1; // plus 1 to store the last column in global memory
const size_t BlockIterations = RefLen % BlockSizeInColumns==0?RefLen / BlockSizeInColumns:RefLen / BlockSizeInColumns + 1;
const size_t LastBlockSizeInColumns= RefLen%BlockIterations==0?0:RefLen - ((BlockIterations-1)*BlockSizeInColumns);

const size_t GlobalBarrierLen = QueNum * QueLenMax * RefPadBloLen;

//std::cout << "Block Iterations for RefLen - block size "<<BlockIterations<<" for "<<RefLen<<" with block size "<<BlockSizeInColumns<<" and last block size in columns: "<<LastBlockSizeInColumns<<"\n";
//std::cout<<"Global Mem Size: "<<GlobalBarrierSize<<"\n";
//const size_t gMemSAddr = get_sGloIndex (SharedMemSize, RefPadBloLen*3);
//const size_t SharedMemSizePerMP = gMemSAddr;
//const size_t GlobalMemSizePerMP = InterRefVecLen - SharedMemSizePerMP;
//const size_t TotalGlobalMemSize = QueNum * GlobalMemSizePerMP;

//std::cout << "Total/ Global/ Shared"<< InterRefVecLen <<"/"<< GlobalMemSizePerMP <<"/" << SharedMemSizePerMP <<"\n";


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = InterRefVecLen;
ArgVec[5] = QueLenMax;
//ArgVec[0] = RefLen;
//ArgVec[1] = QueNum;
//ArgVec[2] = RefPadBloLen;
for(size_t i = 6; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-6].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;

//std::cout<<"Before/After:"<<IntVecLen<<"/"<<InterIntVecLen*4/1024/1024<<"\n";

const size_t ResVecLen = QueNum * RefPadBloLen;


// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);

cl_mem GlobalBarrierSDDev = allocateDeviceMemory_u32 (GlobalBarrierLen, myContext);

cl_mem GlobalBarrierSLDev = allocateDeviceMemory_u32 (GlobalBarrierLen, myContext);

cl_mem GlobalBarrierSCDev = allocateDeviceMemory_u32 (GlobalBarrierLen, myContext);

cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clSetKernelArg(myKernel, 5, sizeof(unsigned int) * 7, 0);
clSetKernelArg(myKernel, 6, sizeof(unsigned int) * SharedMemSize32,0);
clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &GlobalBarrierSDDev);
clSetKernelArg(myKernel, 8, sizeof(cl_mem), (void *) &GlobalBarrierSLDev);
clSetKernelArg(myKernel, 9, sizeof(cl_mem), (void *) &GlobalBarrierSCDev);

clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/

/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-REFERENCE GLOBAL - BLOCKED 2

timer NikosTimer;
const size_t RefNum = 128;
const size_t QueNum = 200;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen;
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}

// SIZE CHECK AND FIT TO SHARED AND GLOBAL MEMORY // TODO:NOW

const size_t SharedLineSize = 5; // KBytes
const size_t SharedLineLen = (SharedLineSize * 1024) / 4 ;
const size_t SharedSiteNum =  SharedLineLen / RefPadBloLen - 1;
const size_t BLIters = RefLen / SharedSiteNum;
std::cout<< "BLiters: " << BLIters <<" Sites: "<<SharedSiteNum<<"\n";
const size_t BlockIterNum = RefLen % SharedSiteNum==0?(RefLen / SharedSiteNum): (RefLen / SharedSiteNum) + 1;

const size_t GlobalBarrierLen = QueNum * QueLenMax * RefPadBloLen;

//std::cout << "Block Iterations for RefLen - block size "<<BlockIterNum<<" for "<<RefLen<<" with block size "<<SharedSiteNum<<" and last block size in columns: "<<SharedSiteNum*BlockIterNum<<"\n";



// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgCount = 8;
const size_t ArgVecLen = QueNum + ArgCount; 
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

unsigned int ind=0;
ArgVec[ind++] = QueNum;
ArgVec[ind++] = RefNum;
ArgVec[ind++] = RefLen;
ArgVec[ind++] = RefPadBloLen;
ArgVec[ind++] = InterRefVecLen;
ArgVec[ind++] = QueLenMax;
ArgVec[ind++] = BLIters;//SharedSiteNum;
ArgVec[ind++] = BlockIterNum;

for(size_t i = ind; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-ind].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;

//std::cout<<"Before/After:"<<IntVecLen<<"/"<<InterIntVecLen*4/1024/1024<<"\n";

const size_t ResVecLen = QueNum * RefPadBloLen;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem GlobalBarrierSDDev = allocateDeviceMemory_u32 (GlobalBarrierLen, myContext);

cl_mem GlobalBarrierSLDev = allocateDeviceMemory_u32 (GlobalBarrierLen, myContext);

cl_mem GlobalBarrierSCDev = allocateDeviceMemory_u32 (GlobalBarrierLen, myContext);



cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);

clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &GlobalBarrierSDDev);
clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &GlobalBarrierSLDev);
clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &GlobalBarrierSCDev);
clSetKernelArg(myKernel, 8, sizeof(unsigned int) * SharedLineLen,0);

clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);

std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/




/******************************************************************************************************************************
// ** WORKING IMPLEMENTATION:  INTER-REFERENCE GLOBAL 

timer NikosTimer;
const size_t RefNum = 190; //  
const size_t QueNum = 200;
const size_t RefLen = m_ref_pvecs[0].size();
unsigned int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;

std::cout << "Run Load: "<< RefNum << " references " << QueNum << " queries " << RefLen << " reference length and " << QueLenMax << " max query length \n\n";

// STEP 1 
// ORGANIZE QUERIES // Simple placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax;
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}

// STEP 2 
// ORGANIZE REFERENCES  // Simple placement one after the other in blocks of size RefLen.
const size_t RefVecLen = RefNum * RefLen;
unsigned int * RefVec = (unsigned int *) malloc(sizeof(unsigned int)*RefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,RefVec);}

// ORGANIZE REFERENCES  // Inter-reference organization in multiples of 32.
const size_t RefPadBloLen = RefNum;//get_next32mult (RefNum);
const size_t InterRefVecLen = RefPadBloLen * RefLen;
unsigned int * InterRefVec = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<InterRefVecLen;i++){InterRefVec[i]=0;}
for(size_t i=0;i<RefNum;i++){putSequenceRef_Inter(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec,RefPadBloLen);}

// SIZE CHECK AND FIT TO SHARED AND GLOBAL MEMORY // TODO IN NEXT STEP
//const size_t SharedMemSize = 48; // KBytes
//const size_t gMemSAddr = get_sGloIndex (SharedMemSize, RefPadBloLen*3);
//const size_t SharedMemSizePerMP = gMemSAddr;
//const size_t GlobalMemSizePerMP = InterRefVecLen - SharedMemSizePerMP;
//const size_t TotalGlobalMemSize = QueNum * GlobalMemSizePerMP;

//std::cout << "Total/ Global/ Shared"<< InterRefVecLen <<"/"<< GlobalMemSizePerMP <<"/" << SharedMemSizePerMP <<"\n";


// STEP 3 
// INITIALIZE ARGUMENT VECTOR 
const size_t ArgVecLen = QueNum + 6; // Set to RefLen for kernel testing
unsigned int * ArgVec = (unsigned int *) malloc(sizeof(unsigned int)*ArgVecLen);

// Temp initialization for testing

ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = InterRefVecLen;
ArgVec[4] = QueLenMax;
//ArgVec[0] = RefLen;
//ArgVec[1] = QueNum;
//ArgVec[2] = RefPadBloLen;
for(size_t i = 5; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-5].size();
}

const size_t IntVecLen = QueNum * RefNum * RefLen; 

const size_t InterIntVecLen = QueNum *  InterRefVecLen;

//std::cout<<"Before/After:"<<IntVecLen<<"/"<<InterIntVecLen*4/1024/1024<<"\n";

const size_t ResVecLen = QueNum * RefPadBloLen;

// STEP 4 
// MEMORY ALLOCATION, INITIALIZATION, and TRANSFER TO DEVICE 

cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem RefVecPin = allocatePinnedHostMemory_u32 (RefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (RefVecLen, RefVecPin, myCommandQueue, RefVec);
cl_mem RefVecDev = allocateDeviceMemory_u32 (RefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (RefVecLen, RefVecPin, RefVecDev, myCommandQueue);

cl_mem InterRefVecPin = allocatePinnedHostMemory_u32 (InterRefVecLen, myContext, myCommandQueue);
RefVecPin = initializePinnedHostMemory_u32_Sequence (InterRefVecLen, InterRefVecPin, myCommandQueue, InterRefVec);
cl_mem InterRefVecDev = allocateDeviceMemory_u32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (InterRefVecLen, InterRefVecPin, InterRefVecDev, myCommandQueue);


cl_mem ArgVecPin = allocatePinnedHostMemory_u32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_u32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_u32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem IntVecDev = allocateDeviceMemory_u32 (IntVecLen, myContext);
cl_mem InterIntVecDev = allocateDeviceMemory_u32 (InterIntVecLen, myContext);

cl_mem ResVecPin = allocatePinnedHostMemory_u32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_u32 (ResVecLen, myContext);

clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &RefVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
//clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &IntVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &InterIntVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &ResVecDev);
clFinish(myCommandQueue);

// STEP 5 
// LAUNCH KERNEL 
//size_t WorkSizeGlobal[] = {RefNum * QueNum};
size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
//size_t WorkSizeLocal[] = {RefNum};//argumentsSize};
size_t WorkSizeLocal[] = {RefPadBloLen};
cl_event event;

clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);

clFinish(myCommandQueue);

// STEP 6 
// READ DATA BACK AND PRINT 
unsigned int * ResVec = copyDeviceMemoryToPinnedHostMemory_u32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);
double NikosTime= NikosTimer.elapsed();
std::cout << "GPU Time: " << NikosTimer.elapsed();

/*****/














/* SIMON'S REFERENCE TEST */

unsigned int score_ref;
//for(size_t i=0;i<1;i++)
//{
//score_ref = align_pvec_score( m_ref_pvecs[1], m_ref_aux[1], m_qs_pvecs[198], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
//std::cout << "Simon score for 0 0 0     " << score_ref <<"\n";
//}
timer SimonTimerN;
unsigned int * ResVecRef = (unsigned int *) malloc (sizeof(unsigned int)*ResVecLen);
size_t y = 0, yy=0;
unsigned int fillzeros = RefPadBloLen - RefNum;
unsigned int ncups = 0;
align_arrays arr;
for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) {
 for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) {          


                
                unsigned int score = align_pvec_score( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		ncups += RefLen * m_qs_pvecs[i].size();
		ResVecRef[y++] = score;	
		
		//std::cout<<i<<" "<<j<<" "<<score<<"\n";

            }  
	    for(size_t yy=0;yy<fillzeros;yy++)
	    {
		ResVecRef[y++] = 0;
 	    }          
            
        }

std::cout << "\nCPU Time: " << SimonTimerN.elapsed() <<"\n\n";
unsigned int totalcounter=0;

//std::cout << "GNcupsCPU: " << ncups / (NikosTime * 1e9) << "\n";
//std::cout << "GNcupsCPU: " << ncups / (SimonTimerN.elapsed() * 1e9) << "\n";
for(size_t i=0;i<ResVecLen;i++)
{
	
	if(ResVec[i]!=ResVecRef[i])
	{
		std::cout<<i<<":"<<ResVec[i]<<" and " << ResVecRef[i]<<"\n";
		totalcounter++;
	}
}

std::cout<<"\n\nIt is:"<<totalcounter<<" out of:"<<ResVecLen<<"\n\n";
/**/





//unsigned int * ReferenceSequenceReturned = copyDeviceMemoryToPinnedHostMemory_u32 (compReferenceLength, referenceDevice, referencePinned, myCommandQueue);
//unsigned int * QuerySequenceReturned = copyDeviceMemoryToPinnedHostMemory_u32 (queryLength, queryDevice, queryPinned, myCommandQueue);

//print_sequence(ReferenceSequenceReturned, compReferenceLength);
//print_sequence(QuerySequenceReturned, queryLength);

//int score_s2 = align_pvec_score_experimental3(ReferenceSequenceReturned,QuerySequenceReturned,compReferenceLength,queryLength, score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );

//std::cout << "\n\nNikos score for 0 0 0     " << score_s2 << "\n\n\n\n";


/******************************************/






	/*******************************************************************************************************/
	// Experiments with GPU performance

	//printf("\n\nCompact Size per Ref (u32): %d\n",compactRefSize); // Size of Reference Sequence in u32

/*
unsigned int * refReturned = NULL;	
timer myTimer;




for(size_t i=0;i<1000;i++)
{
	
	
	cl_mem refPinned = allocatePinnedHostMemory_u32 (compactRefSize, myContext, myCommandQueue); // allocate pinned memory on host

	refPinned = initializePinnedHostMemory_u32 (compactRefSize, refPinned, myCommandQueue); // initialize pinned memory
	//refPinned2 = initializePinnedHostMemory_u32 (compactRefSize, refPinned2, myCommandQueue); // initialize pinned memory


	cl_mem refDevice = allocateDeviceMemory_u32 (compactRefSize, myContext); // allocate device memory
	//cl_mem refDevice2 = allocateDeviceMemory_u32 (compactRefSize, myContext); // allocate device memory

	copyPinnedHostMemoryToDeviceMemory_u32 (compactRefSize, refPinned, refDevice, myCommandQueue); // transfer data
	//copyPinnedHostMemoryToDeviceMemory_u32 (compactRefSize, refPinned2, refDevice2, myCommandQueue); // transfer data
  


	clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &refDevice);
	//clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &refDevice2);

	clFinish(myCommandQueue);


        /*
        size_t workGroupSize[1];
	    
	workGroupSize[0] = compactRefSize;
	    
	cl_event event;

	clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, workGroupSize, NULL, 0, NULL, &event);
	
	clFinish(myCommandQueue);

	/*


        refReturned = copyDeviceMemoryToPinnedHostMemory_u32 (compactRefSize, refDevice, refPinned, myCommandQueue);


	clFinish(myCommandQueue);
}
//std::cout << "\nTime passed:"<<myTimer.elapsed()<<"\n\n";



        /**
	for(size_t i=0; i < compactRefSize; i++)
	{
		printf("%d:%u\n",i,refReturned[i]);
		
	}
        /**/



/*** Dumb Implementation ***/






/***************************/

/**
//timer t3;
	for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) {
            
            //const size_t reflen = m_ref_pvecs[i].size();

            for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) {
                

                
               
		int score2 = align_pvec_score_experimental( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		 int score = align_pvec_score( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
                 
		 if( score != score2 ) {
                    std::cout << "score " << i << " " << j << " = " << score << "\n";
                 }
                ncup += reflen * m_qs_pvecs[i].size();
            }
            
            
        }
//std::cerr << "aligned " << m_ref_pvecs.size() << " x " << m_qs_pvecs.size() << " seqs in " << t1.elapsed() << " s\n";
/**
timer t1;
	for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) {
            
            //const size_t reflen = m_ref_pvecs[i].size();

            for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) {
                

                
               
		int score2 = align_pvec_score_experimental( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		 int score = score2;//align_pvec_score( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
                 
		 if( score != score2 ) {
                    std::cout << "score " << i << " " << j << " = " << score << "\n";
                 }
                ncup += reflen * m_qs_pvecs[i].size();
            }
            
            
        }
std::cerr << "aligned " << m_ref_pvecs.size() << " x " << m_qs_pvecs.size() << " seqs in " << t1.elapsed() << " s\n";
/**
timer t2;
/**
	// Reference run implemented by Simon
        

        for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) {
            
            const size_t reflen = m_ref_pvecs[i].size();
            std::cerr << "ref: " << i << "\n";
            for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) {
                

                
                int score = align_pvec_score( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		int score2 = score;

                 if( score != score2 ) {
                    std::cout << "score " << i << " " << j << " = " << score << "\n";
                 }
                ncup += reflen * m_qs_pvecs[i].size();
            }
            
            
        }
      /**/  
       // std::cerr << "aligned " << m_ref_pvecs.size() << " x " << m_qs_pvecs.size() << " seqs in " << t2.elapsed() << " s\n";
        
    //    std::cerr << "GNcups: " << ncup / (t1.elapsed() * 1e9) << "\n";
        
    }
};



int main() {
    testbench tb( "qs.bin", "ref_new.bin" );
    tb.run_seq();
}
